//
//  ChallongeServer.swift
//  Amalgamation
//
//  Created by Charles Magahern on 10/8/17.
//

import Foundation

class ChallongeServer
{
    private var _urlSession:        URLSession
    private var _operationQueue:    OperationQueue
    
    fileprivate static let _baseURL = URL(string: "https://api.challonge.com/v1/")!
    fileprivate static let _apiKey = "???"
    
    init()
    {
        let config = URLSessionConfiguration.default
        _urlSession = URLSession(configuration: config)
        
        _operationQueue = OperationQueue()
        _operationQueue.maxConcurrentOperationCount = 1
    }
    
    func loadTournament(tournamentStringID: String,
                        loadParticipants: Bool,
                        loadMatches: Bool,
                        completion: @escaping (Tournament?, Error?) -> Void)
    {
        let operation = LoadTournamentOperation(baseURL: ChallongeServer._baseURL, session: _urlSession)
        operation.tournamentStringID = tournamentStringID
        operation.loadParticipants = loadParticipants
        operation.loadMatches = loadMatches
        
        weak var weakOp = operation
        operation.completionBlock = {
            guard let strongOp = weakOp else { completion(nil, nil) ; return }
            if let error = strongOp.error {
                self._logError("failed to load tournament", error: error)
            }
            
            completion(strongOp.tournament, strongOp.error)
        }
        
        _operationQueue.addOperation(operation)
    }
    
    func loadTournamentParticipants(tournament: Tournament, completion: @escaping ([Participant]?, Error?) -> Void)
    {
        let operation = LoadParticipantsOperation(baseURL: ChallongeServer._baseURL, session: _urlSession)
        operation.tournamentStringID = tournament.stringID
        
        weak var weakOp = operation
        operation.completionBlock = {
            guard let strongOp = weakOp else { completion(nil, nil) ; return }
            if let error = strongOp.error {
                self._logError("failed to load participants", error: error)
            }
            
            completion(strongOp.participants, strongOp.error)
        }
        
        _operationQueue.addOperation(operation)
    }
    
    func loadTournamentMatches(tournament: Tournament, completion: @escaping ([Match]?, Error?) -> Void)
    {
        let operation = LoadMatchesOperation(baseURL: ChallongeServer._baseURL, session: _urlSession)
        operation.tournamentStringID = tournament.stringID
        
        weak var weakOp = operation
        operation.completionBlock = {
            guard let strongOp = weakOp else { completion(nil, nil) ; return }
            if let error = strongOp.error {
                self._logError("failed to load matches", error: error)
            }
            
            completion(strongOp.matches, strongOp.error)
        }
        
        _operationQueue.addOperation(operation)
    }
    
    // MARK: Internal
    
    fileprivate class func _jsonDecoder() -> JSONDecoder
    {
        let dateFormatter = DateFormatter()
        dateFormatter.locale = Locale(identifier: "en_US_POSIX")
        dateFormatter.dateFormat = "yyyy-MM-dd'T'HH:mm:ss.SSSZZZZZ"
        dateFormatter.timeZone = TimeZone(secondsFromGMT: 0)
        
        let decoder = JSONDecoder()
        decoder.dateDecodingStrategy = .formatted(dateFormatter)
        
        return decoder
    }
    
    internal func _logError(_ description: String, error: Error)
    {
        #if DEBUG
            StandardErrorOutputStream.shared.write("ERROR: \(description) \(error.localizedDescription)\n")
        #endif
    }
}

internal class LoadTournamentOperation : ServerOperation
{
    var tournamentStringID: String = ""
    var loadParticipants: Bool = false
    var loadMatches: Bool = false
    internal(set) var tournament: Tournament?
    
    override func main()
    {
        let url = self.baseURL.appendingPathComponent("tournaments/\(self.tournamentStringID).json")
        var urlComponents = URLComponents(url: url, resolvingAgainstBaseURL: false)!
        urlComponents.query = "api_key=\(ChallongeServer._apiKey)&include_participants=\(self.loadParticipants ? 1 : 0)&include_matches=\(self.loadMatches ? 1 : 0)"
        
        if let data = self.fetchData(urlComponents.url!) {
            let decoder = ChallongeServer._jsonDecoder()
            
            do {
                let response = try decoder.decode([String : Tournament].self, from: data)
                self.tournament = response["tournament"]
            } catch {
                // try to read error from server
                do {
                    let response = try decoder.decode([String : [String]].self, from: data)
                    let errorMsg = response["errors"]!.first!
                    if errorMsg == "Requested tournament not found" {
                        self.error = ChallongeError(.tournamentNotFound)
                    } else {
                        self.error = ChallongeError(.unknown)
                    }
                } catch {
                    self.error = ChallongeError(.parseFailure)
                }
            }
        }
    }
}

internal class LoadParticipantsOperation : ServerOperation
{
    var tournamentStringID: String = ""
    internal(set) var participants: [Participant]?
    
    override func main()
    {
        let url = self.baseURL.appendingPathComponent("tournaments/\(self.tournamentStringID)/participants.json")
        var urlComponents = URLComponents(url: url, resolvingAgainstBaseURL: false)!
        urlComponents.query = "api_key=\(ChallongeServer._apiKey)"
        
        if let data = self.fetchData(urlComponents.url!) {
            do {
                let decoder = ChallongeServer._jsonDecoder()
                let response = try decoder.decode([[String : Participant]].self, from: data)
                self.participants = response.map({$0["participant"]!})
            } catch {
                self.error = ChallongeError(.parseFailure)
            }
        }
    }
}

internal class LoadMatchesOperation : ServerOperation
{
    var tournamentStringID: String = ""
    internal(set) var matches: [Match]?
    
    override func main()
    {
        let url = self.baseURL.appendingPathComponent("tournaments/\(self.tournamentStringID)/matches.json")
        var urlComponents = URLComponents(url: url, resolvingAgainstBaseURL: false)!
        urlComponents.query = "api_key=\(ChallongeServer._apiKey)"
        
        if let data = self.fetchData(urlComponents.url!) {
            do {
                let decoder = ChallongeServer._jsonDecoder()
                let response = try decoder.decode([[String : Match]].self, from: data)
                self.matches = response.map({$0["match"]!})
            } catch {
                self.error = ChallongeError(.parseFailure)
            }
        }
    }
}
