//
//  Tournament.swift
//  Amalgamation
//
//  Created by Charles Magahern on 10/8/17.
//

import Foundation

struct Tournament : Decodable
{
    enum State : String, Codable
    {
        case unknown
        case pending
        case underway
        case complete
        
        var localizedDescription: String
        {
            switch self {
            case .pending:
                return NSLocalizedString("TOURNAMENT_STATE_PENDING", comment: "")
            case .underway:
                return NSLocalizedString("TOURNAMENT_STATE_UNDERWAY", comment: "")
            case .complete:
                return NSLocalizedString("TOURNAMENT_STATE_COMPLETE", comment: "")
            default:
                return NSLocalizedString("UNKNOWN", comment: "")
            }
        }
    }
    
    enum TournamentType : String, Codable
    {
        case unknown
        case singleElimination = "single elimination"
        case doubleElimination = "double elimination"
        case roundRobin = "round robin"
        case swiss
        
        var localizedDescription: String
        {
            switch self {
            case .singleElimination:
                return NSLocalizedString("TOURNAMENT_TYPE_SINGLE_ELIM", comment: "")
            case .doubleElimination:
                return NSLocalizedString("TOURNAMENT_TYPE_DOUBLE_ELIM", comment: "")
            case .roundRobin:
                return NSLocalizedString("TOURNAMENT_TYPE_ROUND_ROBIN", comment: "")
            case .swiss:
                return NSLocalizedString("TOURNAMENT_TYPE_SWISS", comment: "")
            default:
                return NSLocalizedString("UNKNOWN", comment: "")
            }
        }
    }
    
    var id:                 Int            = 0
    var stringID:           String         = ""
    var state:              State          = .unknown
    var type:               TournamentType = .unknown
    
    var dateCreated:        Date    = Date()
    var dateStarted:        Date?
    var dateCompleted:      Date?
    var dateUpdated:        Date?
    
    var description:        String  = ""
    var gameID:             Int     = 0
    var gameName:           String  = ""
    var name:               String  = ""
    var participantsCount:  UInt    = 0
    var challongeURL:       URL?
    var liveImageURL:       URL?
    
    // maps participant/player IDs to participant objects
    var participantsMap:    [Int : Participant]
    
    // complete array of matches for this tournament
    var matches:            [Match]
    
    // MARK: Coding Support
    
    enum CodingKeys : String, CodingKey
    {
        case id
        case stringID = "url"
        case state
        case type = "tournament_type"
        case dateCreated = "created_at"
        case dateStarted = "started_at"
        case dateCompleted = "completed_at"
        case dateUpdated = "updated_at"
        case description
        case gameID = "game_id"
        case gameName = "game_name"
        case name
        case participantsCount = "participants_count"
        case challongeURL = "full_challonge_url"
        case liveImageURL = "live_image_url"
        case participants
        case matches
    }
    
    init(from decoder: Decoder) throws
    {
        let container = try decoder.container(keyedBy: CodingKeys.self)
        self.id = try container.decode(Int.self, forKey: .id)
        self.stringID = try container.decode(String.self, forKey: .stringID)
        self.state = try container.decode(State.self, forKey: .state)
        self.type = try container.decode(TournamentType.self, forKey: .type)
        self.dateCreated = try container.decode(Date.self, forKey: .dateCreated)
        self.dateStarted = try container.decodeIfPresent(Date.self, forKey: .dateStarted)
        self.dateCompleted = try container.decodeIfPresent(Date.self, forKey: .dateCompleted)
        self.dateUpdated = try container.decodeIfPresent(Date.self, forKey: .dateUpdated)
        self.description = try container.decode(String.self, forKey: .description)
        self.gameID = try container.decodeIfPresent(Int.self, forKey: .gameID) ?? 0
        self.gameName = try container.decodeIfPresent(String.self, forKey: .gameName) ?? "<Unknown>"
        self.name = try container.decode(String.self, forKey: .name)
        self.participantsCount = try container.decode(UInt.self, forKey: .participantsCount)
        self.challongeURL = try container.decodeIfPresent(URL.self, forKey: .challongeURL)
        self.liveImageURL = try container.decodeIfPresent(URL.self, forKey: .liveImageURL)
        
        if let participantsContainer = try container.decodeIfPresent([[String : Participant]].self, forKey: .participants) {
            self.participantsMap = participantsContainer.reduce([Int : Participant](), { (inDict, parsedParticipantDict) -> [Int : Participant] in
                let participant = parsedParticipantDict["participant"]!
                var pair = inDict
                pair[participant.id] = participant
                return pair
            })
        } else {
            self.participantsMap = [Int : Participant]()
        }
        
        if let matchesContainer = try container.decodeIfPresent([[String : Match]].self, forKey: .matches) {
            self.matches = matchesContainer.map({$0["match"]!})
        } else {
            self.matches = []
        }
    }
    
    //! returns sorted array of (participantID, matchesCount) pairs, sorted from
    //! most matches to fewest matches
    func participantsSortedByMatchCount() -> [(key: Int, value: UInt)]
    {
        var playerIDToMatchesPlayedMap: [Int : UInt] = [:]
        for match in self.matches {
            guard let player1ID = match.player1ID else { continue }
            guard let player2ID = match.player2ID else { continue }
            
            let player1CurrentCount = playerIDToMatchesPlayedMap[player1ID] ?? 0
            playerIDToMatchesPlayedMap[player1ID] = player1CurrentCount + 1
            
            let player2CurrentCount = playerIDToMatchesPlayedMap[player2ID] ?? 0
            playerIDToMatchesPlayedMap[player2ID] = player2CurrentCount + 1
        }
        
        let sortedMap = playerIDToMatchesPlayedMap.sorted(by: { $0.value > $1.value })
        return sortedMap
    }
}
