//
//  DDRCardDrawServer.swift
//  Amalgamation
//
//  Created by Charles Magahern on 9/22/18.
//

import Foundation

struct DDRCardDrawError : Error
{
    enum Code
    {
        case unknown
        case connectionFailure
        case noMatchAvailable
        case parseFailure
    }
    
    let code: Code
    
    init(_ code: Code)
    {
        self.code = code
    }
}

class DDRCardDrawServer
{
    private var _urlSession:        URLSession
    private var _operationQueue:    OperationQueue
    
    fileprivate static let _baseURL = URL(string: "http://mephisto.zanneth.com:5730")!
    
    init()
    {
        let config = URLSessionConfiguration.default
        _urlSession = URLSession(configuration: config)
        
        _operationQueue = OperationQueue()
        _operationQueue.maxConcurrentOperationCount = 1
    }
    
    func fetchCurrentMatch(completion: @escaping (DDRMatch?, Error?) -> Void)
    {
        let operation = LoadDDRMatchOperation(baseURL: DDRCardDrawServer._baseURL, session: _urlSession)
        
        weak var weakOp = operation
        operation.completionBlock = {
            guard let strongOp = weakOp else { completion(nil, nil) ; return }
            if let error = strongOp.error {
                self._logError("failed to load matches", error: error)
            }
            
            completion(strongOp.ddrMatch, strongOp.error)
        }
        
        _operationQueue.addOperation(operation)
    }
    
    // MARK: Internal
    
    internal func _logError(_ description: String, error: Error)
    {
        #if DEBUG
            StandardErrorOutputStream.shared.write("ERROR: \(description) \(error.localizedDescription)\n")
        #endif
    }
}

internal class LoadDDRMatchOperation : ServerOperation
{
    internal(set) var ddrMatch: DDRMatch?
    
    override func main()
    {
        let urlRequest = URLRequest(url: self.baseURL)
        let (data, response) = self.fetchResponse(urlRequest)
        
        var errorCode: DDRCardDrawError.Code?
        if let httpResponse = response as? HTTPURLResponse {
            switch (httpResponse.statusCode) {
            case 200:
                errorCode = nil
            case 404:
                errorCode = .noMatchAvailable
            default:
                errorCode = .unknown
            }
        } else {
            errorCode = .connectionFailure
        }
        
        if let errorCode = errorCode {
            self.error = DDRCardDrawError(errorCode)
        } else {
            do {
                let decoder = JSONDecoder()
                let jsonObject = try decoder.decode([String : DDRMatch].self, from: data!)
                self.ddrMatch = jsonObject["match"]
            } catch {
                self.error = DDRCardDrawError(.parseFailure)
            }
        }
    }
}
