//
//  DDRCardDrawViewController.swift
//  Amalgamation
//
//  Created by Charles Magahern on 9/21/18.
//

import Foundation
import UIKit

class DDRCardDrawViewController : UIViewController, CustomTournamentScreen, Transitionable
{
    private let _server:               DDRCardDrawServer = DDRCardDrawServer()
    private var _cachedUpdate:         DDRMatch?
    
    private let _kanjiLabel:           UILabel = UILabel()
    private let _playerOneLabel:       UILabel = UILabel()
    private let _playerTwoLabel:       UILabel = UILabel()
    private let _versusLabel:          UILabel = UILabel()
    
    private var _cardViews:            [DDRCardView] = []
    private var _transitionInProgress: Bool = false
    private var _active:               Bool = false
    
    // MARK: API
    
    var ddrMatch: DDRMatch?
    {
        didSet
        {
            _reloadPlayersUI()
            _reloadCardViews()
        }
    }
    
    func reloadData()
    {
        _server.fetchCurrentMatch { (match: DDRMatch?, error: Error?) in
            DispatchQueue.main.async {
                self._cachedUpdate = match
                
                if !self._transitionInProgress {
                    self._reloadFromCachedResult()
                }
            }
        }
    }
    
    // MARK: UIViewController
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        let theme = Theme.mainTheme
        let headerFont = theme.headerTextAttributes[NSAttributedStringKey.font] as! UIFont
        let centeredParagraphStyle = NSMutableParagraphStyle()
        centeredParagraphStyle.alignment = .center
        
        let kanjiFontDescriptor = headerFont.fontDescriptor.withSymbolicTraits(.traitBold)!
        let kanjiFont = UIFont(descriptor: kanjiFontDescriptor, size: 72.0)
        let kanjiAttributes: TextAttributes = [
            NSAttributedStringKey.paragraphStyle : centeredParagraphStyle,
            NSAttributedStringKey.font : kanjiFont,
            NSAttributedStringKey.foregroundColor : UIColor.white
        ]
        _kanjiLabel.attributedText = NSAttributedString(string: "試合状態", attributes: kanjiAttributes)
        
        let view = self.view
        view?.addSubview(_kanjiLabel)
        view?.addSubview(_playerOneLabel)
        view?.addSubview(_playerTwoLabel)
        view?.addSubview(_versusLabel)
    }
    
    override func viewDidLayoutSubviews()
    {
        super.viewDidLayoutSubviews()
        
        let bounds = self.view.bounds
        let kanjiPlayerNamesVPadding = CGFloat(20.0)
        let playerUICardsVPadding = bounds.size.height / 5.0
        
        // pre-compute sizes
        let kanjiLabelSize = _kanjiLabel.sizeThatFits(bounds.size)
        let collinearLabels = [_playerOneLabel, _versusLabel, _playerTwoLabel]
        var playerLabelsBoundingSize = CGSize(width: 0.0, height: 0.0)
        
        for label in collinearLabels {
            let labelSize = label.sizeThatFits(bounds.size)
            playerLabelsBoundingSize.width += labelSize.width
            playerLabelsBoundingSize.height = max(playerLabelsBoundingSize.height, labelSize.height)
        }
        
        let cardViewsWidth = rint(bounds.size.width / 9.0)
        let cardViewsSize = CGSize(
            width: cardViewsWidth,
            height: rint(cardViewsWidth * 1.61803) // golden ratio
        )
        
        let viewsTotalHeight = (
            kanjiLabelSize.height +
            kanjiPlayerNamesVPadding +
            playerLabelsBoundingSize.height +
            playerUICardsVPadding +
            cardViewsSize.height
        )
        let viewsOriginY = rint(bounds.origin.y + (bounds.size.height / 2.0 - viewsTotalHeight / 2.0))
        
        // layout kanji label
        let kanjiLabelFrame = CGRect(
            x: rint(bounds.origin.x + (bounds.size.width / 2.0 - kanjiLabelSize.width / 2.0)),
            y: viewsOriginY,
            width: kanjiLabelSize.width,
            height: kanjiLabelSize.height
        )
        _kanjiLabel.frame = kanjiLabelFrame
        
        // layout player labels
        let labelsHPadding = CGFloat(8.0)
        let labelsOriginX = rint(bounds.origin.x + (bounds.size.width / 2.0 - playerLabelsBoundingSize.width / 2.0))
        var previousLabel: UILabel?
        
        for label in collinearLabels {
            let labelSize = label.sizeThatFits(bounds.size)
            let originX = ((previousLabel != nil) ? previousLabel!.frame.maxX + labelsHPadding : labelsOriginX)
            let labelFrame = CGRect(
                x: originX,
                y: kanjiLabelFrame.maxY + kanjiPlayerNamesVPadding,
                width: labelSize.width,
                height: labelSize.height
            )
            label.frame = labelFrame
            previousLabel = label
        }
        
        // layout card views
        let cardViewsHPadding = CGFloat(20.0)
        let cardViewsCount = CGFloat(_cardViews.count)
        let cardViewsTotalWidth = (cardViewsCount * cardViewsSize.width) + ((cardViewsCount - 1.0) * cardViewsHPadding)
        let cardViewsOriginX = rint(bounds.origin.x + (bounds.size.width / 2.0 - cardViewsTotalWidth / 2.0))
        let cardViewsOriginY = rint(previousLabel!.frame.maxY + playerUICardsVPadding)
        
        var lastCardViewFrame: CGRect?
        for cardView in _cardViews {
            let originX = ((lastCardViewFrame != nil) ? lastCardViewFrame!.maxX + cardViewsHPadding : cardViewsOriginX)
            let cardViewFrame = CGRect(
                x: originX,
                y: cardViewsOriginY,
                width: cardViewsSize.width,
                height: cardViewsSize.height
            )
            
            cardView.frame = cardViewFrame
            lastCardViewFrame = cardViewFrame
        }
    }
    
    // MARK: CustomTournamentScreen
    
    var shouldShowScreen: Bool
    {
        get
        {
            if (self.ddrMatch == nil) {
                self.reloadData()
            }
            return (self.ddrMatch != nil)
        }
    }
        
    // MARK: Transitionable
    
    func prepareToTransitionToActive()
    {
        self.reloadData()
    }
    
    func transitionToActive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _transitionToActive(active: true, transition: transition, completion: completion)
    }
    
    func transitionToInactive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _transitionToActive(active: false, transition: transition, completion: completion)
    }
    
    // MARK: Internal
    
    internal func _reloadPlayersUI()
    {
        let unknownPlayerString = NSLocalizedString("MATCH_UNKNOWN_PLAYER_NAME", comment: "")
        let playerOneName = (self.ddrMatch?.playerOneName ?? unknownPlayerString).uppercased()
        let playerTwoName = (self.ddrMatch?.playerTwoName ?? unknownPlayerString).uppercased()
        
        let theme = Theme.mainTheme
        var versusTextAttributes = theme.headerTextAttributes
        let headerTextFont = versusTextAttributes[NSAttributedStringKey.font] as! UIFont
        let boldHeaderTextFontDesc = headerTextFont.fontDescriptor.withSymbolicTraits(.traitBold)!
        versusTextAttributes[NSAttributedStringKey.font] = UIFont(descriptor: boldHeaderTextFontDesc, size: 55.0)
        
        var playerNameAttributes = versusTextAttributes
        playerNameAttributes[NSAttributedStringKey.foregroundColor] = theme.colorPalette.primaryColor
        
        _playerOneLabel.attributedText = NSAttributedString(string: playerOneName, attributes: playerNameAttributes)
        _playerTwoLabel.attributedText = NSAttributedString(string: playerTwoName, attributes: playerNameAttributes)
        _versusLabel.attributedText = NSAttributedString(string: "vs.", attributes: versusTextAttributes)
        
        self.view.setNeedsLayout()
    }
    
    internal func _reloadCardViews()
    {
        _cardViews.forEach { $0.removeFromSuperview() }
        _cardViews.removeAll()
        
        let songs = self.ddrMatch?.cardDraws ?? []
        for song in songs {
            let cardView = DDRCardView()
            cardView.song = song
            cardView.layer.opacity = (_active ? 1.0 : 0.0)
            
            _cardViews.append(cardView)
            self.view.addSubview(cardView)
        }
        
        self.view.setNeedsLayout()
    }
    
    internal func _reloadFromCachedResult()
    {
        self.ddrMatch = _cachedUpdate
    }
    
    internal func _inTransaction_transitionPlayersUI(active: Bool, transition: Transition)
    {
        let playersUILabels = [_kanjiLabel, _playerOneLabel, _playerTwoLabel, _versusLabel]
        playersUILabels.forEach { $0.layer.removeAllAnimations() }
        
        let playerLabelsOffset = CGFloat(20.0)
        var animations: [CAAnimation] = []
        
        let fadeAnimation = CABasicAnimation(keyPath: "opacity")
        fadeAnimation.fromValue = CGFloat(active ? 0.0 : 1.0)
        fadeAnimation.toValue = CGFloat(active ? 1.0 : 0.0)
        animations.append(fadeAnimation)
        
        let playerOneAnimation = CABasicAnimation(keyPath: "transform.translation.x")
        playerOneAnimation.fromValue = (active ? -playerLabelsOffset : 0.0)
        playerOneAnimation.toValue = (active ? 0.0 : -playerLabelsOffset)
        animations.append(playerOneAnimation)
        
        let playerTwoAnimation = CABasicAnimation(keyPath: "transform.translation.x")
        playerTwoAnimation.fromValue = (active ? playerLabelsOffset : 0.0)
        playerTwoAnimation.toValue = (active ? 0.0 : playerLabelsOffset)
        animations.append(playerTwoAnimation)
        
        playersUILabels.forEach { $0.layer.opacity = (active ? 0.0 : 1.0) }
        
        // set common properties on the animations
        for animation in animations {
            animation.duration = transition.duration
            animation.beginTime = CACurrentMediaTime() + transition.delay
            animation.timingFunction = CAMediaTimingFunction(name: (active ? kCAMediaTimingFunctionEaseOut : kCAMediaTimingFunctionEaseIn))
        }
        
        playersUILabels.forEach { $0.layer.add(fadeAnimation, forKey: nil) }
        _playerOneLabel.layer.add(playerOneAnimation, forKey: nil)
        _playerTwoLabel.layer.add(playerTwoAnimation, forKey: nil)
        
        DispatchQueue.main.asyncAfter(deadline: DispatchTime.secondsFromNow(transition.delay)) {
            playersUILabels.forEach { $0.layer.opacity = (active ? 1.0 : 0.0) }
        }
    }
    
    internal func _inTransaction_transitionCardViews(active: Bool, transition: Transition)
    {
        _cardViews.forEach { $0.layer.removeAllAnimations() }
        
        let cardViewsOffset = CGFloat(100.0)
        let timingFunction = CAMediaTimingFunction(name: (active ? kCAMediaTimingFunctionEaseOut : kCAMediaTimingFunctionEaseIn))
        var animations: [CAAnimation] = []
        
        let fadeAnimation = CABasicAnimation(keyPath: "opacity")
        fadeAnimation.fromValue = CGFloat(active ? 0.0 : 1.0)
        fadeAnimation.toValue = CGFloat(active ? 1.0 : 0.0)
        fadeAnimation.duration = transition.duration
        fadeAnimation.timingFunction = timingFunction
        fadeAnimation.fillMode = kCAFillModeForwards
        animations.append(fadeAnimation)
        
        let transformAnimation = CABasicAnimation(keyPath: "transform.translation.y")
        transformAnimation.fromValue = (active ? cardViewsOffset : 0.0)
        transformAnimation.toValue = (active ? 0.0 : cardViewsOffset)
        transformAnimation.duration = transition.duration
        transformAnimation.timingFunction = timingFunction
        transformAnimation.fillMode = kCAFillModeForwards
        animations.append(transformAnimation)
        
        _cardViews.forEach { $0.layer.opacity = (active ? 0.0 : 1.0) }
        
        let delayStagger = 0.1
        var delayOffset = 0.0
        for cardView in _cardViews {
            let beginTime = CACurrentMediaTime() + transition.delay + delayOffset
            
            let cardTransformAnimation = transformAnimation.copy() as! CABasicAnimation
            cardTransformAnimation.beginTime = beginTime
            
            let cardFadeAnimation = fadeAnimation.copy() as! CABasicAnimation
            cardFadeAnimation.beginTime = beginTime
            
            cardView.layer.add(cardFadeAnimation, forKey: nil)
            cardView.layer.add(cardTransformAnimation, forKey: nil)
            
            delayOffset += delayStagger
            
            DispatchQueue.main.asyncAfter(deadline: DispatchTime.secondsFromNow(transition.delay + delayOffset)) {
                cardView.layer.opacity = (active ? 1.0 : 0.0)
            }
        }
    }
    
    internal func _transitionToActive(active: Bool, transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _transitionInProgress = true
        
        CATransaction.begin()
        CATransaction.setCompletionBlock {
            completion(true)
            
            self._active = active
            self._transitionInProgress = false
            self._reloadFromCachedResult()
        }
        
        _inTransaction_transitionPlayersUI(active: active, transition: transition)
        _inTransaction_transitionCardViews(active: active, transition: transition)
        
        CATransaction.commit()
    }
}
