//
//  DDRCardView.swift
//  Amalgamation
//
//  Created by Charles Magahern on 9/21/18.
//

import Foundation
import UIKit

class DDRCardView : UIView
{
    private let _containerView:     UIView = UIView()
    private let _titleLabel:        UILabel = UILabel()
    private let _artistLabel:       UILabel = UILabel()
    private let _bpmLabel:          UILabel = UILabel()
    private let _difficultyLabel:   UILabel = UILabel()
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        self.addSubview(_containerView)
        
        let textAttributes = Theme.mainTheme.bodyTextAttributes
        let font = textAttributes[NSAttributedStringKey.font] as! UIFont
        let textColor = textAttributes[NSAttributedStringKey.foregroundColor] as! UIColor
        
        // set common properties, add labels as subviews
        for label in [_titleLabel, _artistLabel, _bpmLabel, _difficultyLabel] {
            label.font = font
            label.textColor = textColor
            
            _containerView.addSubview(label)
        }
        
        let boldFontDescriptor = font.fontDescriptor.withSymbolicTraits(.traitBold)!
        _titleLabel.numberOfLines = 2
        _titleLabel.font = UIFont(descriptor: boldFontDescriptor, size: 28.0)
        _titleLabel.textAlignment = .center
        
        _artistLabel.numberOfLines = 2
        _artistLabel.font = font.withSize(22.0)
        _artistLabel.textAlignment = .center
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: Accessors
    
    var song: DDRSong?
    {
        didSet
        {
            _titleLabel.text = song?.title
            _artistLabel.text = song?.artist
            _bpmLabel.text = String(format: "%@ BPM", (song?.bpm ?? "?"))
            _difficultyLabel.text = song?.chart?.difficultyDescription
            
            _containerView.backgroundColor = song?.chart?.difficultyColor
            
            if song?.vetoed ?? false {
                _containerView.alpha = 0.5
            } else {
                _containerView.alpha = 1.0
            }
            
            self.setNeedsLayout()
        }
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let edgePadding = CGFloat(5.0)
        let titleArtistPadding = CGFloat(5.0)
        
        _containerView.frame = bounds
        
        let titleLabelSize = _titleLabel.sizeThatFits(bounds.size)
        let artistLabelSize = _artistLabel.sizeThatFits(bounds.size)
        let titleArtistTotalHeight = (
            titleLabelSize.height +
            titleArtistPadding +
            artistLabelSize.height
        )
        
        let titleLabelFrame = CGRect(
            x: rint(bounds.size.width / 2.0 - titleLabelSize.width / 2.0),
            y: rint(bounds.size.height / 2.0 - titleArtistTotalHeight / 2.0),
            width: titleLabelSize.width,
            height: titleLabelSize.height
        )
        _titleLabel.frame = titleLabelFrame
        
        let artistLabelFrame = CGRect(
            x: rint(bounds.size.width / 2.0 - artistLabelSize.width / 2.0),
            y: titleLabelFrame.maxY + titleArtistPadding,
            width: artistLabelSize.width,
            height: artistLabelSize.height
        )
        _artistLabel.frame = artistLabelFrame
        
        let bpmLabelSize = _bpmLabel.sizeThatFits(bounds.size)
        let bpmLabelFrame = CGRect(
            x: edgePadding,
            y: rint(bounds.size.height - edgePadding - bpmLabelSize.height),
            width: bpmLabelSize.width,
            height: bpmLabelSize.height
        )
        _bpmLabel.frame = bpmLabelFrame
        
        let difficultyLabelSize = _difficultyLabel.sizeThatFits(bounds.size)
        let difficultyLabelFrame = CGRect(
            x: rint(bounds.size.width - edgePadding - difficultyLabelSize.width),
            y: rint(bounds.size.height - edgePadding - bpmLabelSize.height),
            width: difficultyLabelSize.width,
            height: difficultyLabelSize.height
        )
        _difficultyLabel.frame = difficultyLabelFrame
    }
}

extension DDRChart
{
    var difficultyDescription: String
    {
        get
        {
            var difficultyKey: String!
            
            switch (self.difficulty) {
            case .beginner:
                difficultyKey = "DDR_DIFFICULTY_SHORT_BEGINNER"
            case .basic:
                difficultyKey = "DDR_DIFFICULTY_SHORT_BASIC"
            case .difficult:
                difficultyKey = "DDR_DIFFICULTY_SHORT_DIFFICULT"
            case .expert:
                difficultyKey = "DDR_DIFFICULTY_SHORT_EXPERT"
            case .challenge:
                difficultyKey = "DDR_DIFFICULTY_SHORT_CHALLENGE"
            }
            
            let difficultyName = NSLocalizedString(difficultyKey, comment: "")
            return String(format: "%d %@", self.rating, difficultyName)
        }
    }
    
    var difficultyColor: UIColor
    {
        get
        {
            var color: UIColor!
            
            switch (self.difficulty) {
            case .beginner:
                color = #colorLiteral(red: 0.1411764771, green: 0.3960784376, blue: 0.5647059083, alpha: 1)
            case .basic:
                color = #colorLiteral(red: 0.7254902124, green: 0.4784313738, blue: 0.09803921729, alpha: 1)
            case .difficult:
                color = #colorLiteral(red: 0.4392156899, green: 0.01176470611, blue: 0.1921568662, alpha: 1)
            case .expert:
                color = #colorLiteral(red: 0.2745098174, green: 0.4862745106, blue: 0.1411764771, alpha: 1)
            case .challenge:
                color = #colorLiteral(red: 0.3647058904, green: 0.06666667014, blue: 0.9686274529, alpha: 1)
            }
            
            return color
        }
    }
}
