//
//  PoissonUniform.swift
//  Amalgamation
//
//  Created by Charles Magahern on 2/4/18.
//

import CoreGraphics
import Foundation

class PoissonUniform
{
    let size:               CGSize
    let minimumDistance:    CGFloat
    var pointsPerSample:    Int = 30
    
    private var _grid:      [[CGPoint]] = []
    private var _cellSize:  CGFloat = 0.0
    
    private static let _maxPoints = 100000
    
    init(size: CGSize, minimumDistance: CGFloat)
    {
        self.size = size
        self.minimumDistance = minimumDistance
        
        _cellSize = self.minimumDistance / sqrt(2.0)
    }
    
    func generatePoints() -> [CGPoint]
    {
        _grid = _createGrid(width: Int(ceil(self.size.width / _cellSize)), height: Int(ceil(self.size.height / _cellSize)))
        
        var samplePoints: [CGPoint] = []
        var processList: [CGPoint] = []
        
        let firstPoint = CGPoint(x: _frand() * self.size.width, y: _frand() * self.size.height)
        processList.append(firstPoint)
        samplePoints.append(firstPoint)
        
        let (gridX, gridY) = _imageToGrid(point: firstPoint)
        _grid[gridY][gridX] = firstPoint
        
        while processList.count > 0 && samplePoints.count < PoissonUniform._maxPoints {
            let randIdx = Int(arc4random()) % processList.count
            let point = processList[randIdx]
            processList.remove(at: randIdx)
            
            for _ in 0 ..< self.pointsPerSample {
                let newPoint = _randomPoint(around: point)
                let shouldAddPoint = (
                    newPoint.x > 0.0 &&
                    newPoint.y > 0.0 &&
                    newPoint.x < self.size.width &&
                    newPoint.y < self.size.height &&
                    !_pointInNeighborhood(point: newPoint)
                )
                
                if shouldAddPoint {
                    processList.append(newPoint)
                    samplePoints.append(newPoint)
                    
                    let (gridX, gridY) = _imageToGrid(point: newPoint)
                    _grid[gridY][gridX] = newPoint
                }
            }
        }
        
        return samplePoints
    }
    
    // MARK: Internal
    
    internal func _frand() -> CGFloat
    {
        return CGFloat(MathUtil.frand())
    }
    
    internal func _createGrid(width: Int, height: Int) -> [[CGPoint]]
    {
        return [[CGPoint]](repeating: [CGPoint](repeating: CGPoint(x: 0.0, y: 0.0), count: width), count: height)
    }
    
    internal func _imageToGrid(point: CGPoint) -> (Int, Int)
    {
        return (
            Int(point.x / _cellSize),
            Int(point.y / _cellSize)
        )
    }
    
    internal func _randomPoint(around point: CGPoint) -> CGPoint
    {
        let r1 = _frand()
        let r2 = _frand()
        let radius = self.minimumDistance + (r1 * self.minimumDistance)
        let angle = 2.0 * CGFloat.pi * r2
        
        return CGPoint(
            x: point.x + radius * CGFloat(cos(angle)),
            y: point.y + radius * CGFloat(sin(angle))
        )
    }
    
    internal func _distance(pt1: CGPoint, pt2: CGPoint) -> CGFloat
    {
        return sqrt(pow(pt2.x - pt1.x, 2.0) + pow(pt2.y - pt1.y, 2.0))
    }
    
    internal func _pointInNeighborhood(point: CGPoint) -> Bool
    {
        let gridPoint = _imageToGrid(point: point)
        
        for i in max(0, gridPoint.0 - 2) ..< min(_grid[0].count, gridPoint.0 + 3) {
            for j in max(0, gridPoint.1 - 2) ..< min(_grid.count, gridPoint.1 + 3) {
                let cell = _grid[j][i]
                
                if _distance(pt1: cell, pt2: point) < self.minimumDistance {
                    return true
                }
            }
        }
        
        return false
    }
}
