//
//  SettingsViewController.swift
//  Amalgamation
//
//  Created by Charles Magahern on 6/2/18.
//

import UIKit

class SettingsViewController : UIViewController, UITableViewDelegate, UITableViewDataSource, TournamentSettingsDelegate
{
    private let _server:                           ChallongeServer = ChallongeServer()
    
    private let _titleLabel:                       UILabel = UILabel()
    private let _subtitleLabel:                    UILabel = UILabel()
    private let _loadingBarView:                   LoadingBarView = LoadingBarView()
    
    private var _splitViewController:              UISplitViewController = UISplitViewController()
    private var _tournamentsTableViewController:   UITableViewController = UITableViewController()
    private var _tournamentSettingsViewController: TournamentSettingsViewController = TournamentSettingsViewController()
    
    private var _focusedTournamentsTableIndexPath: IndexPath?
    private var _savedTournaments:                 [SavedTournament] = []

    static private let _TournamentCellIdentifier    = "TournamentCellIdentifier"
    static private let _SavedTournamentsDefaultsKey = "SavedTournaments"
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        if let view = self.view {
            let theme = Theme.mainTheme
            let themeHeaderFont = theme.headerTextAttributes[NSAttributedStringKey.font] as! UIFont
            let centeredParagraphStyle = NSMutableParagraphStyle()
            centeredParagraphStyle.alignment = .center
            
            // setup view
            view.backgroundColor = theme.backgroundColor

            // setup title view
            let titleFontPointSize = CGFloat(48.0)
            let titleFont = themeHeaderFont.withSize(titleFontPointSize)
            let titleAttributes: TextAttributes = [
                NSAttributedStringKey.paragraphStyle : centeredParagraphStyle,
                NSAttributedStringKey.font : titleFont,
                NSAttributedStringKey.foregroundColor : UIColor.white,
                NSAttributedStringKey.kern : (360.0 * titleFontPointSize / 1000.0)
            ]
            _titleLabel.attributedText = NSAttributedString(string: "AMALGAMATION", attributes: titleAttributes)
            view.addSubview(_titleLabel)
            
            // setup subtitle view
            let subtitleFontDescriptor = titleFont.fontDescriptor.withSymbolicTraits(.traitBold)!
            let subtitleFont = UIFont(descriptor: subtitleFontDescriptor, size: 36.0)
            let subtitleAttributes: TextAttributes = [
                NSAttributedStringKey.paragraphStyle : centeredParagraphStyle,
                NSAttributedStringKey.font : subtitleFont,
                NSAttributedStringKey.foregroundColor : UIColor.white
            ]
            _subtitleLabel.attributedText = NSAttributedString(string: "試合観客制度", attributes: subtitleAttributes)
            view.addSubview(_subtitleLabel)
            
            // setup loading bar
            _setLoadingState(isLoading: false)
            view.addSubview(_loadingBarView)
            
            // setup table view controller
            let tournamentsTableView = _tournamentsTableViewController.tableView!
            tournamentsTableView.delegate = self
            tournamentsTableView.dataSource = self
            tournamentsTableView.remembersLastFocusedIndexPath = true
            tournamentsTableView.register(UITableViewCell.self, forCellReuseIdentifier: SettingsViewController._TournamentCellIdentifier)
            
            // setup tournament settings view controller
            _tournamentSettingsViewController.delegate = self

            // setup split view controller
            _splitViewController.viewControllers = [_tournamentsTableViewController, _tournamentSettingsViewController]
            self.addChildViewController(_splitViewController)
            view.addSubview(_splitViewController.view)
            
            // load data
            _loadSavedTournaments()
            tournamentsTableView.reloadData()
        }
    }
    
    override func viewDidLayoutSubviews()
    {
        super.viewDidLayoutSubviews()
        
        let bounds = self.view.bounds
        let insets = TVConstants.contentSafeZoneInsets
        let loadingBarHeight = CGFloat(20.0)
        
        // layout title label
        let titleSizeThatFits = _titleLabel.sizeThatFits(bounds.size)
        let titleFrame = CGRect(
            x: rint(bounds.size.width / 2.0 - titleSizeThatFits.width / 2.0),
            y: insets.top,
            width: titleSizeThatFits.width,
            height: titleSizeThatFits.height
        )
        _titleLabel.frame = titleFrame
        
        // layout subtitle label
        let subtitleSizeThatFits = _subtitleLabel.sizeThatFits(bounds.size)
        let subtitleFrame = CGRect(
            x: rint(bounds.size.width / 2.0 - subtitleSizeThatFits.width / 2.0),
            y: titleFrame.maxY + 10.0,
            width: subtitleSizeThatFits.width,
            height: subtitleSizeThatFits.height
        )
        _subtitleLabel.frame = subtitleFrame
        
        // layout tournament settings view (split view controller)
        let tournamentSettingsSize = CGSize(
            width: bounds.size.width / 1.5,
            height: bounds.size.height / 1.5
        )
        let tournamentSettingsFrame = CGRect(
            x: rint(bounds.size.width / 2.0 - tournamentSettingsSize.width / 2.0),
            y: rint(subtitleFrame.maxY + (bounds.size.height - subtitleFrame.maxY) / 2.0 - tournamentSettingsSize.height / 2.0),
            width: tournamentSettingsSize.width,
            height: tournamentSettingsSize.height
        )
        _splitViewController.minimumPrimaryColumnWidth = tournamentSettingsSize.width / 3.0
        _splitViewController.view.frame = tournamentSettingsFrame
        
        // layout loading bar
        let loadingBarFrame = CGRect(
            x: tournamentSettingsFrame.origin.x,
            y: bounds.size.height - insets.bottom - loadingBarHeight,
            width: tournamentSettingsFrame.size.width,
            height: loadingBarHeight
        )
        _loadingBarView.frame = loadingBarFrame
    }

    override weak var preferredFocusedView: UIView?
    {
        var preferredFocusedView: UIView? = nil
        
        // if the currently focused table view cell is the "new tournament" cell, then
        // we should be focusing on the tournament settings view
        if let focusedIndexPath = _focusedTournamentsTableIndexPath {
            if focusedIndexPath.row >= _savedTournaments.count {
                preferredFocusedView = _tournamentSettingsViewController.view
            }
        }
        
        return preferredFocusedView
    }
    
    // MARK: UITableViewDataSource
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int
    {
        return _savedTournaments.count + 1 // plus one create cell
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell
    {
        let cell = tableView.dequeueReusableCell(withIdentifier: SettingsViewController._TournamentCellIdentifier, for: indexPath)
        cell.selectionStyle = .none

        if indexPath.row < _savedTournaments.count {
            let savedTournament = _savedTournaments[indexPath.row]
            cell.textLabel?.text = savedTournament.stringID
            cell.textLabel?.textColor = UIColor.white
        } else {
            cell.textLabel?.text = NSLocalizedString("NEW_TOURNAMENT_CELL_TEXT", comment: "")
            cell.textLabel?.textColor = UIColor.gray
        }

        return cell
    }
    
    // MARK: UITableViewDelegate
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath)
    {
        if let selectedSavedTournament = _savedTournaments[safe: indexPath.row] {
            self._setLoadingState(isLoading: true)
            
            let tournamentStringID = selectedSavedTournament.stringID
            _server.loadTournament(tournamentStringID: tournamentStringID,
                                   loadParticipants: true,
                                   loadMatches: true)
            { (tournament: Tournament?, error: Error?) in
                DispatchQueue.main.async {
                    if let tournament = tournament, let navigationController = self.navigationController {
                        let tournamentViewController = TournamentViewController()
                        tournamentViewController.tournament = tournament
                        tournamentViewController.server = self._server
                        
                        let theme = Theme.mainTheme
                        let tournamentViewTransition = Transition(
                            duration: theme.transitionDuration,
                            delay: 0.0,
                            options: [.curveEaseOut, .beginFromCurrentState]
                        )

                        navigationController.pushViewController(tournamentViewController, animated: true)
                        tournamentViewController.transitionToActive(transition: tournamentViewTransition, completion: { (_) in })
                    } else {
                        let alertTitle = NSLocalizedString("TOURNAMENT_LOAD_ERROR_TITLE", comment: "")
                        let alertMessage = error?.localizedDescription ?? NSLocalizedString("UNKNOWN_ERROR", comment: "")
                        let alert = UIAlertController(title: alertTitle, message: alertMessage, preferredStyle: .alert)
                        alert.addAction(UIAlertAction(title: NSLocalizedString("OK", comment: ""), style: .default, handler: { (action: UIAlertAction) in
                            self.dismiss(animated: true, completion: nil)
                        }))
                        self.present(alert, animated: true, completion: nil)
                    }
    
                    self._setLoadingState(isLoading: false)
                }
            }
        } else {
            self.setNeedsFocusUpdate()
        }
    }
    
    func tableView(_ tableView: UITableView,
                   didUpdateFocusIn context: UITableViewFocusUpdateContext,
                   with coordinator: UIFocusAnimationCoordinator)
    {
        if let nextIndexPath = context.nextFocusedIndexPath {
            let savedTournament = _savedTournaments[safe: nextIndexPath.row]
            _tournamentSettingsViewController.savedTournament = savedTournament
            _focusedTournamentsTableIndexPath = nextIndexPath
        }
    }
    
    // MARK: TournamentSettingsDelegate
    
    func tournamentSettingsViewControllerDidInputTournamentID(_ controller: TournamentSettingsViewController,
                                                              tournamentID: String)
    {
        if let focusedIndexPath = _focusedTournamentsTableIndexPath {
            if var selectedSavedTournament = _savedTournaments[safe: focusedIndexPath.row] {
                // this is an existing tournament. update its properties.
                selectedSavedTournament.stringID = tournamentID
                _savedTournaments[focusedIndexPath.row] = selectedSavedTournament
            } else {
                // this is a new tournament. append and reload.
                let newSavedTournament = SavedTournament(stringID: tournamentID)
                _savedTournaments.append(newSavedTournament)
            }
            
            _tournamentsTableViewController.tableView.reloadData()
            _storeSavedTournaments()
        }
    }
    
    func tournamentSettingsViewControllerDidConfirmRemoval(_ controller: TournamentSettingsViewController)
    {
        if let focusedIndexPath = _focusedTournamentsTableIndexPath {
            _savedTournaments.remove(at: focusedIndexPath.row)
            _tournamentsTableViewController.tableView.reloadData()
            _storeSavedTournaments()
            
            self.setNeedsFocusUpdate()
        }
    }

    // MARK: Internal
    
    internal func _loadSavedTournaments()
    {
        let defaults = UserDefaults.standard
        let defaultsKey = SettingsViewController._SavedTournamentsDefaultsKey
        let savedTournamentStringIDs = defaults.stringArray(forKey: defaultsKey) ?? []
        var savedTournaments: [SavedTournament] = []
        
        for stringID in savedTournamentStringIDs {
            savedTournaments.append(SavedTournament(stringID: stringID))
        }
        
        _savedTournaments = savedTournaments
    }
    
    internal func _storeSavedTournaments()
    {
        let defaults = UserDefaults.standard
        let defaultsKey = SettingsViewController._SavedTournamentsDefaultsKey
        let savedTournamentStringIDs = _savedTournaments.map { $0.stringID }
        
        defaults.set(savedTournamentStringIDs, forKey: defaultsKey)
        defaults.synchronize()
    }
    
    internal func _setLoadingState(isLoading: Bool)
    {
        _loadingBarView.isHidden = !isLoading
        
        let tournamentsTableView = _tournamentsTableViewController.tableView!
        tournamentsTableView.isUserInteractionEnabled = !isLoading
        tournamentsTableView.alpha = (isLoading ? 0.5 : 1.0)
        
        let tournamentSettingsView = _tournamentSettingsViewController.view!
        tournamentSettingsView.isUserInteractionEnabled = !isLoading
        tournamentSettingsView.alpha = (isLoading ? 0.5 : 1.0)
        
        if isLoading {
            _loadingBarView.startAnimating()
        } else {
            _loadingBarView.stopAnimating()
        }
    }
}

// -----------------------------------------------------------------------------

protocol TournamentSettingsDelegate: class
{
    func tournamentSettingsViewControllerDidInputTournamentID(_ controller: TournamentSettingsViewController,
                                                              tournamentID: String)
    
    func tournamentSettingsViewControllerDidConfirmRemoval(_ controller: TournamentSettingsViewController)
}

internal class TournamentSettingsViewController : UIViewController
{
    weak var delegate:                  TournamentSettingsDelegate?
    
    private let _tournamentIDTextField: UITextField = UITextField()
    private let _helpTextLabel:         UILabel     = UILabel()
    private let _deleteButton:          UIButton    = UIButton(type: .plain)

    var savedTournament: SavedTournament?
    {
        didSet
        {
            _tournamentIDTextField.text = savedTournament?.stringID
            _deleteButton.isEnabled = (savedTournament != nil)
        }
    }
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        if let view = self.view {
            _tournamentIDTextField.placeholder = NSLocalizedString("TOURNAMENT_ID_PLACEHOLDER", comment: "")
            _tournamentIDTextField.autocorrectionType = .no
            _tournamentIDTextField.autocapitalizationType = .none
            _tournamentIDTextField.addTarget(self, action: #selector(_textFieldValueChanged), for: .editingDidEnd)
            view.addSubview(_tournamentIDTextField)
            
            let theme = Theme.mainTheme
            let centeredParagraphStyle = NSMutableParagraphStyle()
            centeredParagraphStyle.alignment = .center
            
            var helpTextAttributes = theme.bodyTextAttributes
            helpTextAttributes[NSAttributedStringKey.paragraphStyle] = centeredParagraphStyle
            
            let helpTextString = NSLocalizedString("TOURNAMENT_ID_HELP_TEXT", comment: "")
            _helpTextLabel.attributedText = NSAttributedString(string: helpTextString, attributes: helpTextAttributes)
            _helpTextLabel.numberOfLines = 0
            view.addSubview(_helpTextLabel)
            
            _deleteButton.setTitle(NSLocalizedString("REMOVE_BUTTON_TEXT", comment: ""), for: .normal)
            _deleteButton.tintColor = UIColor.red
            _deleteButton.addTarget(self, action: #selector(_deleteButtonClicked), for: .primaryActionTriggered)
            view.addSubview(_deleteButton)
        }
    }
    
    override func viewDidLayoutSubviews()
    {
        super.viewDidLayoutSubviews()
        
        let bounds = self.view.bounds
        let vpadding = CGFloat(20.0)
        
        let tournamentIDFieldSizeThatFits = _tournamentIDTextField.sizeThatFits(bounds.size)
        let helpTextSizeThatFits = _helpTextLabel.sizeThatFits(bounds.size)
        let deleteButtonSizeThatFits = _deleteButton.sizeThatFits(bounds.size)
        let totalViewsHeight = (
            tournamentIDFieldSizeThatFits.height +
            vpadding +
            helpTextSizeThatFits.height +
            vpadding +
            deleteButtonSizeThatFits.height
        )
        
        let tournamentIDFieldFrame = CGRect(
            x: rint(bounds.size.width / 2.0 - tournamentIDFieldSizeThatFits.width / 2.0),
            y: rint(bounds.size.height / 2.0 - totalViewsHeight / 2.0),
            width: tournamentIDFieldSizeThatFits.width,
            height: tournamentIDFieldSizeThatFits.height
        )
        _tournamentIDTextField.frame = tournamentIDFieldFrame
        
        let helpTextFrame = CGRect(
            x: rint(bounds.size.width / 2.0 - helpTextSizeThatFits.width / 2.0),
            y: tournamentIDFieldFrame.maxY + vpadding,
            width: helpTextSizeThatFits.width,
            height: helpTextSizeThatFits.height
        )
        _helpTextLabel.frame = helpTextFrame
        
        let deleteButtonFrame = CGRect(
            x: rint(bounds.size.width / 2.0 - deleteButtonSizeThatFits.width / 2.0),
            y: helpTextFrame.maxY + vpadding,
            width: deleteButtonSizeThatFits.width,
            height: deleteButtonSizeThatFits.height
        )
        _deleteButton.frame = deleteButtonFrame
    }
    
    // MARK: Actions
    
    @objc
    internal func _textFieldValueChanged(_ sender: UITextField)
    {
        if let textValue = _tournamentIDTextField.text {
            delegate?.tournamentSettingsViewControllerDidInputTournamentID(self, tournamentID: textValue)
        }
    }
    
    @objc
    internal func _deleteButtonClicked(_ sender: UIButton)
    {
        if self.savedTournament != nil {
            delegate?.tournamentSettingsViewControllerDidConfirmRemoval(self)
        }
    }
}
