//
//  TournamentViewController.swift
//  Amalgamation
//
//  Created by Charles Magahern on 10/8/17.
//

import UIKit

class TournamentViewController : UIViewController, Transitionable
{
    enum ScreenState
    {
        case screenVisible
        case screenExiting
        case interstitialEntering
        case interstitialExiting
        case screenEntering
    }
    
    private let _tournamentTitleLabel:        UILabel                       = UILabel()
    private let _tournamentStateView:         TournamentStateView           = TournamentStateView()
    private let _tournamentStatsView:         TournamentStatsView           = TournamentStatsView()
    private let _completionProgressView:      CompletionProgressView        = CompletionProgressView()
    private let _competitivePlayersView:      CompetitivePlayersView        = CompetitivePlayersView()
    private let _tournamentBracketView:       TournamentBracketView         = TournamentBracketView()
    private let _visualizationViewController: VisualizationViewController   = VisualizationViewController()
    
    private var _currentScreenCustomView:     UIView?
    
    private var _lastUpdatedTournamentData:   Tournament?
    private var _tournamentDataUpdateTimer:   Timer?
    
    private var _screens:                     [TournamentScreen] = [] // loaded in viewDidLoad()
    private var _currentScreenIndex:          Int = -1
    private var _currentScreenState:          ScreenState = .screenVisible
    private var _screenTimer:                 Timer?

    private var _currentCompetitivePlayers:   [CompetitivePlayer] = []
    private var _highlightedPlayerIndex:      Int = -1
    private var _highlightedPlayersTimer:     Timer?
    
    private static let _screenChangeTimerInterval       = TimeInterval(20.0)
    private static let _highlightedPlayersTimerInterval = TimeInterval(5.0)
    private static let _tournamentDataUpdateInterval    = TimeInterval(10.0)

    // MARK: API
    
    var tournament: Tournament?
    {
        didSet
        {
            if let tournament = tournament {
                // update title label
                let theme = Theme.mainTheme
                _tournamentTitleLabel.attributedText = NSAttributedString(string: tournament.name.uppercased(),
                                                                          attributes: theme.headerTextAttributes)
                
                // update subviews' tournament objects
                _tournamentStateView.tournament = tournament
                _tournamentStatsView.tournament = tournament
                _completionProgressView.tournament = tournament
                _tournamentBracketView.tournament = tournament
                _visualizationViewController.tournament = tournament
                
                // show the latest round in the bracket view
                _tournamentBracketView.scrollToLatestVisibleRound(animated: false)

                // compute competitive players heuristic and store it
                var competitivePlayers: [CompetitivePlayer] = []
                let maxCompetitivePlayers = _competitivePlayersView.maxVisiblePlayers
                let participantIDMatchCountPairs = tournament.participantsSortedByMatchCount()
                for i in 0 ..< min(maxCompetitivePlayers, participantIDMatchCountPairs.count) {
                    let (participantID, matchesCount) = participantIDMatchCountPairs[i]
                    guard let participant = tournament.participantsMap[participantID] else { continue }
                    
                    let competitivePlayer = CompetitivePlayer(participant: participant, matchesCount: matchesCount)
                    competitivePlayers.append(competitivePlayer)
                }
                
                var shouldReloadHighlightedPlayersTimer: Bool = false
                if let highlightedPlayer = _currentCompetitivePlayers[safe: _highlightedPlayerIndex] {
                    shouldReloadHighlightedPlayersTimer = !competitivePlayers.contains(where: { (newPlayer: CompetitivePlayer) -> Bool in
                        return (newPlayer.participant.id == highlightedPlayer.participant.id)
                    })
                }

                _currentCompetitivePlayers = competitivePlayers
                _competitivePlayersView.competitivePlayers = competitivePlayers
                
                if shouldReloadHighlightedPlayersTimer {
                    _resetHighlightedPlayersTimer()
                }
            }
        }
    }
    
    var server: ChallongeServer?
    {
        didSet
        {
            _resetTournamentDataUpdateTimer()
        }
    }
    
    // MARK: UIViewController
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        if let view = self.view {
            // set main background color
            view.backgroundColor = Theme.mainTheme.backgroundColor
            
            // setup state view
            _tournamentStateView.animating = true
            
            // add child view controllers
            self.addChildViewController(_visualizationViewController)
            
            // add subviews
            view.addSubview(_tournamentTitleLabel)
            view.addSubview(_tournamentStateView)
            view.addSubview(_tournamentStatsView)
            view.addSubview(_completionProgressView)
            view.addSubview(_competitivePlayersView)
            view.addSubview(_tournamentBracketView)
            view.addSubview(_visualizationViewController.view)
            
            // initialize screen configurations
            let mainBracketInterstitial = InterstitialView()
            mainBracketInterstitial.text = NSLocalizedString("MAIN_BRACKET_INTERSTITIAL", comment: "")
            mainBracketInterstitial.image = UIImage(named: "MainBracket")?.withRenderingMode(.alwaysTemplate)
            
            let mainBracketScreen = TournamentScreen(
                interstitialView: mainBracketInterstitial,
                content: TournamentScreen.Content.configuration(TournamentScreenConfiguration(
                    visibleMatchesRange: 0...Int.max,
                    interstitialView: mainBracketInterstitial
                ))
            )
            
            let losersBracketInterstitial = InterstitialView()
            losersBracketInterstitial.text = NSLocalizedString("LOSERS_BRACKET_INTERSTITIAL", comment: "")
            losersBracketInterstitial.image = UIImage(named: "LosersBracket")?.withRenderingMode(.alwaysTemplate)
            
            let losersBracketScreen = TournamentScreen(
                interstitialView: losersBracketInterstitial,
                content: TournamentScreen.Content.configuration(TournamentScreenConfiguration(
                    visibleMatchesRange: Int.min...0,
                    interstitialView: losersBracketInterstitial
                ))
            )
            
            let ddrCardDrawInterstitial = InterstitialView()
            ddrCardDrawInterstitial.text = NSLocalizedString("DDR_CARD_DRAW_INTERSTITIAL", comment: "")
            ddrCardDrawInterstitial.image = UIImage(named: "CardDraw")?.withRenderingMode(.alwaysTemplate)
            
            let ddrCardDrawViewController = DDRCardDrawViewController()
            let ddrCardDrawScreen = TournamentScreen(
                interstitialView: ddrCardDrawInterstitial,
                content: TournamentScreen.Content.viewController(ddrCardDrawViewController)
            )
            
            _screens = [mainBracketScreen, losersBracketScreen, ddrCardDrawScreen]
            _setCurrentScreenIndex(index: 0, animated: false)

            // setup tap gesture recognizer for advancing state
            let tapRecognizer = UITapGestureRecognizer()
            tapRecognizer.addTarget(self, action: #selector(TournamentViewController._advanceTapAction))
            tapRecognizer.allowedPressTypes = [NSNumber(value: UIPressType.playPause.hashValue)]
            self.view.addGestureRecognizer(tapRecognizer)
        }
    }
    
    override func viewDidAppear(_ animated: Bool)
    {
        super.viewDidAppear(animated)
        
        UIApplication.shared.isIdleTimerDisabled = true
        
        _resetTournamentScreenAdvanceTimer()
        _resetHighlightedPlayersTimer()
        _resetTournamentDataUpdateTimer()
    }
    
    override func viewDidDisappear(_ animated: Bool)
    {
        super.viewDidDisappear(animated)
        
        UIApplication.shared.isIdleTimerDisabled = false
        
        _screenTimer?.invalidate()
        _highlightedPlayersTimer?.invalidate()
        _tournamentDataUpdateTimer?.invalidate()
    }
    
    override func viewDidLayoutSubviews()
    {
        super.viewDidLayoutSubviews()
        
        let bounds = self.view.bounds
        let insets = TVConstants.contentSafeZoneInsets
        let vpadding: CGFloat = 20.0
        let hmargin: CGFloat = 20.0
        
        // title label
        let titleSizeThatFits = _tournamentTitleLabel.sizeThatFits(bounds.size)
        let titleFrame = CGRect(
            x: insets.left,
            y: insets.top,
            width: titleSizeThatFits.width,
            height: titleSizeThatFits.height
        )
        _tournamentTitleLabel.frame = titleFrame
        
        // state view
        let tournamentStateViewFrame = CGRect(
            x: insets.left,
            y: titleFrame.maxY + vpadding,
            width: 256.0,
            height: 90.0
        )
        _tournamentStateView.frame = tournamentStateViewFrame
        
        // stats view
        let tournamentStatsSizeThatFits = _tournamentStatsView.sizeThatFits(bounds.size)
        let tournamentStatsViewFrame = CGRect(
            x: insets.left,
            y: tournamentStateViewFrame.maxY + vpadding,
            width: tournamentStatsSizeThatFits.width,
            height: tournamentStatsSizeThatFits.height
        )
        _tournamentStatsView.frame = tournamentStatsViewFrame
        
        // completion progress view
        let completionProgressViewFrame = CGRect(
            x: insets.left,
            y: tournamentStatsViewFrame.maxY + vpadding,
            width: tournamentStatsViewFrame.size.width,
            height: tournamentStatsViewFrame.size.height
        )
        _completionProgressView.frame = completionProgressViewFrame
        
        // competitive players view
        let competitivePlayersView = CGRect(
            x: insets.left,
            y: completionProgressViewFrame.maxY + vpadding,
            width: tournamentStatsViewFrame.size.width,
            height: 180.0
        )
        _competitivePlayersView.frame = competitivePlayersView
        
        // visualization view
        let visualizationHeight: CGFloat = 250.0
        let visualizationFrame = CGRect(
            x: insets.left,
            y: bounds.size.height - insets.bottom - visualizationHeight,
            width: tournamentStatsViewFrame.size.width,
            height: visualizationHeight
        )
        _visualizationViewController.view.frame = visualizationFrame
        
        // main bracket view
        let sidebarWidth = _computeSidebarWidth()
        let tournamentBracketViewOriginX: CGFloat = (insets.left + sidebarWidth + hmargin)
        let tournamentBracketViewOriginY: CGFloat = (titleFrame.maxY + vpadding)
        let tournamentBracketViewFrame = CGRect(
            x: tournamentBracketViewOriginX,
            y: tournamentBracketViewOriginY,
            width: bounds.size.width - insets.right - tournamentBracketViewOriginX,
            height: bounds.size.height - insets.bottom - tournamentBracketViewOriginY
        )
        _tournamentBracketView.frame = tournamentBracketViewFrame
        
        // custom screen view container (if exists)
        _currentScreenCustomView?.frame = tournamentBracketViewFrame
        
        // interstitial view (if exists)
        if let currentInterstitialView = _currentInterstitialView {
            let interstitialBoundsSize = CGSize(
                width: tournamentBracketViewFrame.size.width / 4.0,
                height: tournamentBracketViewFrame.size.height / 4.0
            )
            let interstitialViewFrame = CGRect(
                x: rint(tournamentBracketViewOriginX + tournamentBracketViewFrame.size.width / 2.0 - interstitialBoundsSize.width / 2.0),
                y: rint(tournamentBracketViewOriginY + tournamentBracketViewFrame.size.height / 2.0 - interstitialBoundsSize.height / 2.0),
                width: interstitialBoundsSize.width,
                height: interstitialBoundsSize.height
            )
            currentInterstitialView.frame = interstitialViewFrame
        }
    }
    
    // MARK: Transitionable
    
    func transitionToActive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        self.view.layoutIfNeeded()
        
        let sidebarViewsToAnimate = [
            _tournamentStateView,
            _tournamentStatsView,
            _completionProgressView,
            _competitivePlayersView,
            _visualizationViewController.view
        ] as [UIView]
        
        let sidebarWidth = _computeSidebarWidth()
        let beginTransform = CGAffineTransform(translationX: -sidebarWidth, y: 0.0)
        let endTransform = CGAffineTransform.identity
        
        for sidebarView in sidebarViewsToAnimate {
            sidebarView.alpha = 0.0
            sidebarView.transform = beginTransform
        }
        
        _tournamentBracketView.transitionToActive(transition: transition) { (_) in }
        
        let sidebarViewDelayJitter = 0.1
        var curSidebarViewDelay = transition.delay
        
        for sidebarView in sidebarViewsToAnimate {
            UIView.animate(withDuration: transition.duration, delay: curSidebarViewDelay, options: transition.options, animations: {
                sidebarView.alpha = 1.0
                sidebarView.transform = endTransform
            }, completion: nil)
            
            curSidebarViewDelay += sidebarViewDelayJitter
        }
    }
    
    func transitionToInactive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        // not used currently
    }

    // MARK: Internal
    
    internal func _computeSidebarWidth() -> CGFloat
    {
        let sidebarViews = [_tournamentStateView, _tournamentStatsView]
        return sidebarViews.reduce(into: CGFloat(0.0)) { (result: inout CGFloat, view: UIView) in
            result = max(result, view.frame.size.width)
        }
    }

    internal var _currentInterstitialView: InterstitialView? = nil
    {
        willSet
        {
            if _currentInterstitialView != nil {
                _currentInterstitialView?.removeFromSuperview()
            }
        }
        
        didSet
        {
            if let currentInterstitialView = _currentInterstitialView {
                self.view.addSubview(currentInterstitialView)
                self.view.setNeedsLayout()
            }
        }
    }
    
    internal func _loadTestTournament(withContentsOfFile fileURL: URL)
    {
        let data = try! Data(contentsOf: fileURL)
        let dateFormatter = DateFormatter()
        dateFormatter.locale = Locale(identifier: "en_US_POSIX")
        dateFormatter.dateFormat = "yyyy-MM-dd'T'HH:mm:ss.SSSZZZZZ"
        dateFormatter.timeZone = TimeZone(secondsFromGMT: 0)
        
        let decoder = JSONDecoder()
        decoder.dateDecodingStrategy = .formatted(dateFormatter)
        
        let jsonObject = try! decoder.decode([String : Tournament].self, from: data)
        self.tournament = jsonObject["tournament"]
    }
    
    internal func _resetTournamentScreenAdvanceTimer()
    {
        _screenTimer?.invalidate()
        _screenTimer = nil
        
        let interval = TournamentViewController._screenChangeTimerInterval
        _screenTimer = Timer.scheduledTimer(withTimeInterval: interval, repeats: true, block: { [weak self] (_) in
            self?._advanceToNextAvailableTournamentScreen()
        })
    }
    
    internal func _resetHighlightedPlayersTimer()
    {
        _highlightedPlayersTimer?.invalidate()
        
        let interval = TournamentViewController._highlightedPlayersTimerInterval
        let timer = Timer.scheduledTimer(withTimeInterval: interval, repeats: true) { [weak self] (timer: Timer) in
            guard let curPlayerIndex = self?._highlightedPlayerIndex else { return }
            guard let competitivePlayers = self?._currentCompetitivePlayers else { return }
            guard let visualizationViewController = self?._visualizationViewController else { return }
            guard let competitivePlayersView = self?._competitivePlayersView else { return }
            
            let nextPlayerIndex = competitivePlayers.count > 0 ? (curPlayerIndex + 1) % competitivePlayers.count : 0
            if let nextPlayer = competitivePlayers[safe: nextPlayerIndex] {
                visualizationViewController.setShownParticipant(nextPlayer.participant, animated: true)
                competitivePlayersView.setHighlightedParticipant(participant: nextPlayer.participant, animated: true)
    
                self?._highlightedPlayerIndex = nextPlayerIndex
            }
        }
        
        _highlightedPlayersTimer = timer
        _highlightedPlayersTimer?.fire()
    }
    
    internal func _resetTournamentDataUpdateTimer()
    {
        _tournamentDataUpdateTimer?.invalidate()
        
        let interval = TournamentViewController._tournamentDataUpdateInterval
        let timer = Timer.scheduledTimer(withTimeInterval: interval, repeats: true) { [weak self] (timer: Timer) in
            guard let currentTournamentStringID = self?.tournament?.stringID else { return }
            guard let server = self?.server else {
                timer.invalidate()
                return
            }
            
            server.loadTournament(tournamentStringID: currentTournamentStringID,
                                  loadParticipants: true,
                                  loadMatches: true) { (tournament: Tournament?, error: Error?) in
                if let updatedTournament = tournament {
                    self?._lastUpdatedTournamentData = updatedTournament
                } else if let error = error {
                    StandardErrorOutputStream.shared.write("ERROR: \(error.localizedDescription)\n")
                } else {
                    StandardErrorOutputStream.shared.write("Unknown error occurred")
                }
            }
        }
        
        _tournamentDataUpdateTimer = timer
    }
    
    internal func _reloadTournamentDataUsingCache()
    {
        if let cachedTournamentData = self._lastUpdatedTournamentData {
            self.tournament = cachedTournamentData
            self._lastUpdatedTournamentData = nil
        }
    }
    
    internal func _tournamentMatchesCount(inRange range: CountableClosedRange<Int>) -> UInt
    {
        var countInRange: UInt = 0
        
        if let tournament = self.tournament {
            for match in tournament.matches {
                if range.contains(match.round) {
                    countInRange += 1
                }
            }
        }
        
        return countInRange
    }
    
    internal func _shouldShowScreen(_ screen: TournamentScreen) -> Bool
    {
        switch (screen.content) {
        case .configuration(let configuration):
            return (_tournamentMatchesCount(inRange: configuration.visibleMatchesRange) > 0)
        case .viewController(let viewController):
            return viewController.shouldShowScreen
        }
    }
    
    internal func _transitionableForScreen(_ screen: TournamentScreen) -> Transitionable
    {
        switch (screen.content) {
        case .configuration:
            return self._tournamentBracketView
        case .viewController(let viewController):
            return viewController
        }
    }
    
    internal func _transitionOutScreen(screen: TournamentScreen, completion: @escaping (Bool) -> Void)
    {
        let transitionable = _transitionableForScreen(screen)
        let outTransition = Transition(
            duration: Theme.mainTheme.transitionDuration,
            delay: 0.0,
            options: [.curveEaseIn, .beginFromCurrentState]
        )
        
        transitionable.prepareToTransitionToInactive()
        
        _currentScreenState = .screenExiting
        transitionable.transitionToInactive(transition: outTransition, completion: completion)
    }
    
    internal func _prepareScreen(_ screen: TournamentScreen)
    {
        switch (screen.content) {
        case .configuration(let configuration):
            self._currentScreenCustomView?.removeFromSuperview()
            self._tournamentBracketView.isHidden = false
            
            self._tournamentBracketView.visibleMatchesRange = configuration.visibleMatchesRange
            self._tournamentBracketView.scrollToLatestVisibleRound(animated: false)
        
        case .viewController(let viewController):
            self._tournamentBracketView.isHidden = true
        
            if viewController.parent == nil {
                self.addChildViewController(viewController)
            }
            
            self._currentScreenCustomView = viewController.view
            self.view.addSubview(viewController.view)
            self.view.setNeedsLayout()
        }
    }
    
    internal func _transitionInScreen(_ screen: TournamentScreen, setupCallback: @escaping () -> Void)
    {
        let transitionable = _transitionableForScreen(screen)
        
        let inTransition = Transition(
            duration: Theme.mainTheme.transitionDuration,
            delay: 0.0,
            options: [.curveEaseOut, .beginFromCurrentState]
        )
        
        let outTransition = Transition(
            duration: Theme.mainTheme.transitionDuration,
            delay: 0.0,
            options: [.curveEaseIn, .beginFromCurrentState]
        )
        
        _currentInterstitialView = screen.interstitialView
        _currentInterstitialView?.layoutIfNeeded()
        
        _currentScreenState = .interstitialEntering
        _currentInterstitialView?.transitionToActive(transition: inTransition, completion: { (_) in
            // while the interstitial is visible, reload remote data
            self._reloadTournamentDataUsingCache()
            transitionable.prepareToTransitionToActive()
            
            self._currentScreenState = .interstitialExiting
            self._currentInterstitialView?.transitionToInactive(transition: outTransition, completion: { (_) in
                self._currentInterstitialView = nil
                setupCallback()
                
                self._currentScreenState = .screenEntering
                transitionable.transitionToActive(transition: inTransition, completion: { (_) in
                    self._currentScreenState = .screenVisible
                })
            })
        })
    }
    
    internal func _setCurrentScreenIndex(index: Int, animated: Bool)
    {
        guard let nextScreen = _screens[safe: index] else { return }
        let currentScreen = _screens[safe: _currentScreenIndex]
        
        // don't show the next screen if there isn't any content visible
        if !_shouldShowScreen(nextScreen) {
            return
        }
        
        let setupNextScreenBlock = {
            self._prepareScreen(nextScreen)
            self._currentScreenIndex = index
        }
        
        if animated {
            if let currentScreen = currentScreen {
                self._transitionOutScreen(screen: currentScreen) { (_) in
                    self._transitionInScreen(nextScreen, setupCallback: setupNextScreenBlock)
                }
            } else {
                self._transitionInScreen(nextScreen, setupCallback: setupNextScreenBlock)
            }
        } else {
            setupNextScreenBlock()
        }
    }
    
    internal func _advanceToNextAvailableTournamentScreen()
    {
        if _currentScreenState == .screenVisible {
            var nextScreen: TournamentScreen?
            var screenIndex = _currentScreenIndex
            
            repeat {
                screenIndex = (screenIndex + 1) % _screens.count
                nextScreen = _screens[screenIndex]
            } while (nextScreen != nil ? !_shouldShowScreen(nextScreen!) : false)
            
            _setCurrentScreenIndex(index: screenIndex, animated: true)
        }
    }
    
    @objc internal func _advanceTapAction(recognizer: UITapGestureRecognizer)
    {
        if _currentScreenState == .screenVisible {
            _resetTournamentScreenAdvanceTimer()
            _advanceToNextAvailableTournamentScreen()
        }
    }
}
