//
//  VisualizationViewController.swift
//  Amalgamation
//
//  Created by Charles Magahern on 2/4/18.
//

import Foundation
import GameKit
import SceneKit
import UIKit

class VisualizationViewController : UIViewController
{
    private var _scene:             	          SCNScene = SCNScene()
    private var _sceneView:                       SCNView?
    
    private var _cameraNode:        	          SCNNode = SCNNode()
    private var _lightNode:                       SCNNode = SCNNode()
    private var _constellationNode:               SCNNode?
    private var _constellationPointNodes:         [SCNNode] = []
    private var _playerArcsParentNode:            SCNNode?
    
    private var _currentTournamentHash:           UInt = 0
    private var _shownParticipant:                Participant?
    private var _participantIDToPlayerNodeIndex:  [Int : Int] = [Int : Int]()
    
    private static let _constellationNodeSize   = CGFloat(0.2)
    private static let _constellationRadius     = Float(13.0)
    private static let _arcFadeDuration         = TimeInterval(0.2)

    // MARK: Accessors
    
    var tournament: Tournament?
    {
        didSet
        {
            _reloadModel()
            _reloadShownParticipant()
        }
    }
    
    var shownParticipant: Participant?
    {
        get { return _shownParticipant }
        set { self.setShownParticipant(newValue, animated: false) }
    }
    
    func setShownParticipant(_ participant: Participant?, animated: Bool)
    {
        _shownParticipant = participant
        
        if animated {
            let fadeDuration = VisualizationViewController._arcFadeDuration
            let fadeOutAction = SCNAction.fadeOut(duration: fadeDuration)
            let fadeInAction = SCNAction.fadeIn(duration: fadeDuration)
            
            let reloadAndBeginFadeInBlock = {
                self._reloadShownParticipant()
                self._playerArcsParentNode?.opacity = 0.0
                self._playerArcsParentNode?.runAction(fadeInAction)
            }
            
            if let arcsParentNode = _playerArcsParentNode {
                arcsParentNode.runAction(fadeOutAction, completionHandler: reloadAndBeginFadeInBlock)
            } else {
                reloadAndBeginFadeInBlock()
            }
        } else {
            _reloadShownParticipant()
        }
    }
    
    // MARK: UIViewController
    
    override func loadView()
    {
        let view = SCNView(frame: UIScreen.main.bounds, options: nil)
        view.scene = _scene
        view.allowsCameraControl = false
        
        // for performance reasons, let's try to not render to a view with a clear background
        view.backgroundColor = Theme.mainTheme.backgroundColor
        
        _sceneView = view
        self.view = view
    }
    
    override func viewDidLoad()
    {
        super.viewDidLoad()

        _setupCamera()
        _reloadModel()
        _reloadShownParticipant()
    }
    
    override func viewDidAppear(_ animated: Bool)
    {
        super.viewDidAppear(animated)
        _sceneView?.play(nil)
    }
    
    override func viewDidDisappear(_ animated: Bool)
    {
        super.viewDidDisappear(animated)
        _sceneView?.stop(nil)
    }
    
    // MARK: Internal
    
    internal func _setupCamera()
    {
        let camera = SCNCamera()
        
        _cameraNode.camera = camera
        _cameraNode.position = SCNVector3Make(0.0, 0.0, 30.0)
        _scene.rootNode.addChildNode(_cameraNode)
        
        let centerNode = SCNNode()
        centerNode.position = SCNVector3Zero
        _scene.rootNode.addChildNode(centerNode)
        
        let constraint = SCNLookAtConstraint(target: centerNode)
        _cameraNode.constraints = [constraint]
    }
    
    internal enum ArcDrawMode
    {
        case lines
        case cylinders // thicc
    }
    
    internal func _createArcNode(from node1: SCNNode,
                                 to node2: SCNNode,
                                 color: UIColor,
                                 mode: ArcDrawMode) -> SCNNode
    {
        let arcNode = SCNNode()
        let arcQuaternion = MathUtil.shortestArc(pt1: node2.simdPosition, pt2: node1.simdPosition)
        let nextPointBlock = { (t: Float) -> float3 in
            let sq = simd_slerp(node1.simdOrientation, arcQuaternion, t)
            return node1.simdPosition * simd_matrix3x3(sq)
        }
        
        let material = SCNMaterial()
        material.diffuse.contents = color.withAlphaComponent(0.8)
        material.lightingModel = .constant
        
        let segmentsLength = Float(0.5)
        let arcLength = MathUtil.arcLength(position: node1.simdPosition, orientation: node1.simdOrientation, arcQuaternion: arcQuaternion)
        let segmentsCount = UInt(arcLength / segmentsLength)
        
        var cylinderGeometry: SCNCylinder?
        if mode == .cylinders {
            cylinderGeometry = SCNCylinder(radius: VisualizationViewController._constellationNodeSize, height: CGFloat(segmentsLength))
            cylinderGeometry?.firstMaterial = material
        }
        
        for i in 0 ..< segmentsCount {
            let curPt = nextPointBlock(Float(i) / Float(segmentsCount))
            let nextPt = nextPointBlock(Float(i + 1) / Float(segmentsCount))
            
            switch mode {
            case .lines:
                let lineGeometry = SCNGeometry.lineFrom(vector: curPt, toVector: nextPt)
                lineGeometry.firstMaterial = material
                let lineNode = SCNNode(geometry: lineGeometry)
                arcNode.addChildNode(lineNode)
                
            case .cylinders:
                let lookat = MathUtil.computeLookAtMatrix(eye: nextPt, center: curPt, up: float3(0.0, 1.0, 0.0))
                let lookat4 = MathUtil.matrix3ToMatrix4(lookat)
                
                let node = SCNNode(geometry: cylinderGeometry)
                var transform = SCNMatrix4Identity
                transform = SCNMatrix4Rotate(transform, Float.pi / 2.0, 1.0, 0.0, 0.0)
                transform = SCNMatrix4Mult(transform, SCNMatrix4(lookat4))
                transform = SCNMatrix4Translate(transform, curPt.x, curPt.y, curPt.z)
                
                node.transform = transform
                arcNode.addChildNode(node)
            }
        }
        
        return arcNode
    }
    
    internal func _computeTournamentHash() -> UInt
    {
        var participantsHash: UInt = 0
        var matchPairingHash: UInt = 0
        
        // naive hashing for tournament data shown in this visualization
        if let tournament = self.tournament {
            for (participantID, _) in tournament.participantsMap {
                participantsHash += UInt(participantID)
            }
            
            for match in tournament.matches {
                if let p1ID = match.player1ID, let p2ID = match.player2ID {
                    matchPairingHash += UInt(p1ID ^ p2ID)
                }
            }
        }
        
        return (participantsHash ^ matchPairingHash)
    }
    
    internal func _reloadModel()
    {
        let tournamentHash = _computeTournamentHash()
        if tournamentHash != _currentTournamentHash {
            // setup geometry and material to use for the contellation model
            let geometry = SCNSphere(radius: VisualizationViewController._constellationNodeSize)
            let defaultMaterial = SCNMaterial()
            defaultMaterial.diffuse.contents = UIColor(white: 1.0, alpha: 0.5)
    
            // compute fibonacci sphere for the visualization
            let fibonacciSphere = FibonacciSphere(samples: 500, randomize: true)
            let points = fibonacciSphere.generatePoints()
            let r = VisualizationViewController._constellationRadius
    
            // add point nodes from the fibonacci sphere into the scene
            var spherePointNodes: [SCNNode] = []
            let constellationNode = SCNNode()
            constellationNode.position = SCNVector3Zero
    
            for pt in points {
                let position = SCNVector3(
                    pt.x * r,
                    pt.y * r,
                    pt.z * r
                )
    
                let nodeGeometry = geometry.copy() as! SCNGeometry
                nodeGeometry.firstMaterial = defaultMaterial
    
                let node = SCNNode(geometry: nodeGeometry)
                node.position = position
    
                constellationNode.addChildNode(node)
                spherePointNodes.append(node)
            }
    
            // map player IDs to their respective nodes in the constellation
            // also highlight the player nodes with a different material
            var participantIDToPlayerNodeIndex = [Int : Int]()
            if let tournament = self.tournament {
                let participantsCount = Int(tournament.participantsCount)
                let playerNodeIndexes = Array(0 ..< 500).shuffled()[0 ..< participantsCount]
                var participantIterator = tournament.participantsMap.makeIterator()
    
                let theme = Theme.mainTheme
                let playerPointMaterial = SCNMaterial()
                playerPointMaterial.diffuse.contents = theme.colorPalette.primaryColor
    
                for playerNodeIndex in playerNodeIndexes {
                    let node = spherePointNodes[playerNodeIndex]
                    node.geometry?.firstMaterial = playerPointMaterial
    
                    if let participantPair = participantIterator.next() {
                        participantIDToPlayerNodeIndex[participantPair.key] = playerNodeIndex
                    }
                }
            }
    
            // start a slow rotation action
            let rotAngle = 2.0 * Float.pi
            let rotAxis = SCNVector3Make(0.25, 1.75, 1.0)
            let rotationAction = SCNAction.repeatForever(SCNAction.rotate(by: CGFloat(rotAngle), around: rotAxis, duration: 40.0))
            constellationNode.removeAllActions()
            constellationNode.runAction(rotationAction, forKey: nil)
    
            // store result
            _constellationNode?.removeFromParentNode()
            _constellationPointNodes = spherePointNodes
            _constellationNode = constellationNode
            _participantIDToPlayerNodeIndex = participantIDToPlayerNodeIndex
    
            // add constellation to scene
            _scene.rootNode.addChildNode(constellationNode)
    
            // update hash
            _currentTournamentHash = tournamentHash
        }
    }
    
    internal func _reloadShownParticipant()
    {
        _playerArcsParentNode?.removeFromParentNode()
        
        guard let tournament = self.tournament else { return }
        guard let constellationNode = _constellationNode else { return }
        guard let shownParticipant = self.shownParticipant else { return }
        
        // compute the lines that connect each of the nodes
        var connectedPlayerIDPairs: [(Int, Int)] = []
        let arcColor = Theme.mainTheme.colorPalette.primaryColor
        let playerArcsParentNode = SCNNode()
        
        for match in tournament.matches {
            if let p1ID = match.player1ID, let p2ID = match.player2ID {
                if p1ID == shownParticipant.id || p2ID == shownParticipant.id {
                    guard let p1NodeIndex = _participantIDToPlayerNodeIndex[p1ID] else { continue }
                    guard let p2NodeIndex = _participantIDToPlayerNodeIndex[p2ID] else { continue }
                    
                    var connectionAlreadyMade = false
                    for connectedPlayerIDPair in connectedPlayerIDPairs {
                        if ((p1ID == connectedPlayerIDPair.0 && p2ID == connectedPlayerIDPair.1) ||
                            (p1ID == connectedPlayerIDPair.1 && p2ID == connectedPlayerIDPair.0)) {
                            connectionAlreadyMade = true
                        }
                    }
                    
                    if !connectionAlreadyMade {
                        let p1Node = _constellationPointNodes[p1NodeIndex]
                        let p2Node = _constellationPointNodes[p2NodeIndex]
                        let arcNode = _createArcNode(from: p1Node, to: p2Node, color: arcColor, mode: .lines)
                        
                        playerArcsParentNode.addChildNode(arcNode)
                        connectedPlayerIDPairs.append((p1ID, p2ID))
                    }
                }
            }
        }
        
        constellationNode.addChildNode(playerArcsParentNode)
        _playerArcsParentNode = playerArcsParentNode
    }
}
