//
//  CompetitivePlayersView.swift
//  Amalgamation
//
//  Created by Charles Magahern on 2/3/18.
//

import UIKit

struct CompetitivePlayer
{
    var participant:    Participant
    var matchesCount:   UInt
}

extension CompetitivePlayer : Hashable
{
    static func ==(lhs: CompetitivePlayer, rhs: CompetitivePlayer) -> Bool
    {
        return (
            lhs.participant.id == rhs.participant.id &&
            lhs.matchesCount == rhs.matchesCount
        )
    }
    
    public var hashValue: Int
    {
        get
        {
            return (participant.id ^ Int(matchesCount))
        }
    }
}

class CompetitivePlayersView : UIView
{
    private var _headerLabel:               UILabel = UILabel()
    private var _barViews:                  [CompetitivePlayerBarView] = []
    private var _highlightedParticipant:    Participant?
    
    private static let _barViewsVMargin:    CGFloat = 2.0
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        self.backgroundColor = UIColor.clear
        
        let theme = Theme.mainTheme
        let labelText = NSLocalizedString("COMPETITIVE_PLAYERS_LABEL", comment: "").uppercased()
        _headerLabel.attributedText = NSAttributedString(string: labelText, attributes: theme.bodySubheaderAttributes)
        
        self.addSubview(_headerLabel)
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: Accessors
    
    var maxVisiblePlayers: Int = 5
    {
        didSet
        {
            _reloadBarViews()
        }
    }
    
    var competitivePlayers: [CompetitivePlayer] = []
    {
        didSet
        {
            let oldHash = oldValue.reduce(0) { $0 + $1.hashValue }
            let newHash = competitivePlayers.reduce(0) { $0 + $1.hashValue }
            
            if oldHash != newHash {
                _reloadBarViews()
            }
        }
    }
    
    var highlightedParticipant: Participant?
    {
        get { return _highlightedParticipant }
        set { self.setHighlightedParticipant(participant: newValue, animated: false) }
    }
    
    func setHighlightedParticipant(participant: Participant?, animated: Bool)
    {
        _highlightedParticipant = participant
        _reloadHighlightedPlayer(animated: animated)
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let barChartMargin: CGFloat = 10.0
        
        let headerLabelSizeThatFits = _headerLabel.sizeThatFits(bounds.size)
        let headerLabelFrame = CGRect(
            x: 0.0,
            y: 0.0,
            width: headerLabelSizeThatFits.width,
            height: headerLabelSizeThatFits.height
        )
        _headerLabel.frame = headerLabelFrame
        
        let barChartOriginY = headerLabelFrame.maxY + barChartMargin
        let barViewVMargin = CompetitivePlayersView._barViewsVMargin
        let availableBarViewsHeight = bounds.size.height - barChartOriginY
        let totalSpacingHeight = CGFloat(_barViews.count - 1) * barViewVMargin
        let barViewHeight = rint((availableBarViewsHeight - totalSpacingHeight) / CGFloat(_barViews.count))
        var curOriginY = barChartOriginY
        
        for barView in _barViews {
            let barViewFrame = CGRect(
                x: 0.0,
                y: curOriginY,
                width: bounds.size.width,
                height: barViewHeight
            )
            barView.frame = barViewFrame
            curOriginY += barViewHeight + barViewVMargin
        }
    }
    
    // MARK: Internal
    
    internal func _reloadBarViews()
    {
        _barViews.forEach { $0.removeFromSuperview() }
        _barViews.removeAll()
        
        let sortedPlayers = self.competitivePlayers.sorted { $0.matchesCount > $1.matchesCount }
        let highestMatchesPlayed = sortedPlayers.first?.matchesCount ?? 0
        let theme = Theme.mainTheme
        let colorPalette = theme.colorPalette.colors
        let matchesLabelColor = theme.bodySubheaderAttributes[NSAttributedStringKey.foregroundColor] as! UIColor
        
        for i in 0 ..< min(sortedPlayers.count, self.maxVisiblePlayers) {
            let competitivePlayer = sortedPlayers[i]
            
            let playerBarView = CompetitivePlayerBarView()
            playerBarView.participant = competitivePlayer.participant
            playerBarView.playerNameTextColor = theme.foregroundColor
            playerBarView.barPercentageFill = Float(competitivePlayer.matchesCount) / Float(highestMatchesPlayed)
            playerBarView.barColor = colorPalette[i % colorPalette.count]
            playerBarView.matchesCount = competitivePlayer.matchesCount
            playerBarView.matchCountTextColor = matchesLabelColor
            
            self.addSubview(playerBarView)
            _barViews.append(playerBarView)
        }
        
        self.setNeedsLayout()
    }
    
    internal func _reloadHighlightedPlayer(animated: Bool)
    {
        CATransaction.begin()
        CATransaction.setDisableActions(!animated)
        
        for barView in _barViews {
            if let highlightedParticipant = _highlightedParticipant,
                let barViewParticipant = barView.participant {
                
                barView.highlighted = (barViewParticipant == highlightedParticipant)
            } else {
                barView.highlighted = false
            }
        }
        
        CATransaction.commit()
    }
}

internal class CompetitivePlayerBarView : UIView
{
    private var _barLayer:          CAShapeLayer = CAShapeLayer()
    private var _playerNameLabel:   UILabel = UILabel()
    private var _matchesCountLabel: UILabel = UILabel()
    
    private static let _labelPadding: CGFloat = 4.0
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        _barLayer.fillColor = self.barColor.cgColor
        
        let theme = Theme.mainTheme
        let font = theme.accessoryTextAttributes[NSAttributedStringKey.font] as! UIFont
        _playerNameLabel.textColor = self.playerNameTextColor
        _playerNameLabel.font = font
        
        _matchesCountLabel.textColor = self.matchCountTextColor
        _matchesCountLabel.font = font
        
        self.backgroundColor = UIColor.clear
        self.layer.addSublayer(_barLayer)
        self.addSubview(_playerNameLabel)
        self.addSubview(_matchesCountLabel)
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: Accessors
    
    var barColor: UIColor = UIColor.white
    {
        didSet
        {
            _reloadBarColor()
        }
    }
    
    var barPercentageFill: Float = 1.0
    {
        didSet
        {
            self.setNeedsLayout()
        }
    }
    
    var participant: Participant?
    {
        didSet
        {
            _playerNameLabel.text = participant?.name.uppercased() ?? ""
            self.setNeedsLayout()
        }
    }
    
    var playerNameTextColor: UIColor = UIColor.black
    {
        didSet
        {
            _playerNameLabel.textColor = self.playerNameTextColor
        }
    }
    
    var matchesCount: UInt = 0
    {
        didSet
        {
            _matchesCountLabel.text = "\(self.matchesCount)"
            self.setNeedsLayout()
        }
    }
    
    var matchCountTextColor: UIColor = UIColor.darkGray
    {
        didSet
        {
            _matchesCountLabel.textColor = self.matchCountTextColor
        }
    }
    
    var highlighted: Bool = false
    {
        didSet
        {
            _reloadBarColor()
        }
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let barFrame = CGRect(
            x: 0.0,
            y: 0.0,
            width: rint(CGFloat(self.barPercentageFill) * bounds.size.width),
            height: bounds.size.height
        )
        _barLayer.path = CGPath(rect: barFrame, transform: nil)
        _barLayer.frame = barFrame
        
        let labelPadding = CompetitivePlayerBarView._labelPadding
        let playerNameSizeThatFits = _playerNameLabel.sizeThatFits(bounds.size)
        let playerNameFrame = CGRect(
            x: labelPadding,
            y: rint(bounds.size.height / 2.0 - playerNameSizeThatFits.height / 2.0),
            width: playerNameSizeThatFits.width,
            height: playerNameSizeThatFits.height
        )
        _playerNameLabel.frame = playerNameFrame
        
        let matchesCountSizeThatFits = _matchesCountLabel.sizeThatFits(bounds.size)
        let matchesCountFrame = CGRect(
            x: barFrame.size.width - matchesCountSizeThatFits.width - labelPadding,
            y: rint(bounds.size.height / 2.0 - matchesCountSizeThatFits.height / 2.0),
            width: matchesCountSizeThatFits.width,
            height: matchesCountSizeThatFits.height
        )
        _matchesCountLabel.frame = matchesCountFrame
    }
    
    override func sizeThatFits(_ size: CGSize) -> CGSize
    {
        var sizeThatFits = _playerNameLabel.sizeThatFits(size)
        sizeThatFits.width += CompetitivePlayerBarView._labelPadding * 2.0
        sizeThatFits.height += CompetitivePlayerBarView._labelPadding * 2.0
        
        return sizeThatFits
    }
    
    // MARK: Internal
    
    internal func _reloadBarColor()
    {
        var barColor = self.barColor
        if self.highlighted {
            var h, s, b, a: CGFloat; h = 0.0; s = 0.0; b = 0.0; a = 0.0
            barColor.getHue(&h, saturation: &s, brightness: &b, alpha: &a)
            barColor = UIColor(hue: h, saturation: s, brightness: min(b + 0.2, 1.0), alpha: a)
        }
        
        _barLayer.fillColor = barColor.cgColor
    }
}
