//
//  CompletionProgressView.swift
//  Amalgamation
//
//  Created by Charles Magahern on 2/3/18.
//

import UIKit

class CompletionProgressView : UIView
{
    private var _headerLabel:   UILabel = UILabel()
    private var _pieChartView:  PieChartView = PieChartView()
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        self.backgroundColor = UIColor.clear
        
        let theme = Theme.mainTheme
        let labelText = NSLocalizedString("TOURNAMENT_COMPLETION_LABEL", comment: "").uppercased()
        _headerLabel.attributedText = NSAttributedString(string: labelText,
                                                         attributes: theme.bodySubheaderAttributes)
        
        _pieChartView.completedProgressColor = theme.colorPalette.primaryColor
        _pieChartView.incompleteProgressColor = theme.colorPalette.secondaryColor
        
        self.addSubview(_headerLabel)
        self.addSubview(_pieChartView)
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: Accessors
    
    var tournament: Tournament?
    {
        didSet
        {
            if let tournament = self.tournament {
                let matchesCount = tournament.matches.count
                let completedMatchesCount = tournament.matches.reduce(0, { (currentCount: UInt, match: Match) -> UInt in
                    return currentCount + (match.state == .complete ? 1 : 0)
                })
                _pieChartView.percentage = Float(completedMatchesCount) / Float(matchesCount)
            }
        }
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let pieChartMargin: CGFloat = 10.0
        
        let headerLabelSizeThatFits = _headerLabel.sizeThatFits(bounds.size)
        let headerLabelFrame = CGRect(
            x: 0.0,
            y: 0.0,
            width: headerLabelSizeThatFits.width,
            height: headerLabelSizeThatFits.height
        )
        _headerLabel.frame = headerLabelFrame
        
        let pieChartOriginY = headerLabelFrame.maxY + pieChartMargin
        let pieChartFrame = CGRect(
            x: 0.0,
            y: pieChartOriginY,
            width: bounds.size.width,
            height: bounds.size.height - pieChartOriginY
        )
        _pieChartView.frame = pieChartFrame
    }
}

internal class PieChartView : UIView
{
    private var _percentageLabel: UILabel = UILabel()
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        self.backgroundColor = UIColor.clear
        self.addSubview(_percentageLabel)
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: Accessors
    
    var percentage: Float = 0.0
    {
        didSet
        {
            let theme = Theme.mainTheme
            let percentageString = String(format: "%.0f%%", self.percentage * 100.0)
            var attributes = theme.statusTextAttributes
            attributes[NSAttributedStringKey.foregroundColor] = UIColor.white
            attributes[NSAttributedStringKey.backgroundColor] = UIColor.clear
            
            _percentageLabel.attributedText = NSAttributedString(string: percentageString, attributes: attributes)
            
            self.setNeedsLayout()
            self.setNeedsDisplay()
        }
    }
    
    var completedProgressColor: UIColor = UIColor.white
    {
        didSet
        {
            self.setNeedsDisplay()
        }
    }
    
    var incompleteProgressColor: UIColor = UIColor.darkGray
    {
        didSet
        {
            self.setNeedsDisplay()
        }
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let chartRadius = min(bounds.size.width, bounds.size.height) / 2.0
        
        let percentageLabelSizeThatFits = _percentageLabel.sizeThatFits(bounds.size)
        let percentageLabelFrame = CGRect(
            x: rint(chartRadius - percentageLabelSizeThatFits.width / 2.0),
            y: rint(chartRadius - percentageLabelSizeThatFits.height / 2.0),
            width: percentageLabelSizeThatFits.width,
            height: percentageLabelSizeThatFits.height
        )
        _percentageLabel.frame = percentageLabelFrame
    }
    
    override func draw(_ rect: CGRect)
    {
        super.draw(rect)
        
        let bounds = self.bounds
        let radius = rint(min(bounds.size.width, bounds.size.height) / 2.0)
        let progressRadians = CGFloat(self.percentage) * (2.0 * CGFloat.pi)
        let center = CGPoint(x: radius, y: radius)
        
        // fill complete path
        let completedPath = UIBezierPath()
        let completedPathArcStartAngle = (3.0 / 2.0) * CGFloat.pi
        let completedPathArcEndAngle = completedPathArcStartAngle + progressRadians
        completedPath.move(to: center)
        completedPath.addArc(withCenter: center,
                             radius: radius,
                             startAngle: completedPathArcStartAngle,
                             endAngle: completedPathArcEndAngle,
                             clockwise: true)
        completedPath.close()
        
        self.completedProgressColor.setFill()
        completedPath.fill()
        
        // fill incomplete path
        let incompletePath = UIBezierPath()
        incompletePath.move(to: center)
        incompletePath.addArc(withCenter: center,
                              radius: radius,
                              startAngle: completedPathArcEndAngle,
                              endAngle: (3.0 * CGFloat.pi) / 2.0,
                              clockwise: true)
        incompletePath.close()
        
        self.incompleteProgressColor.setFill()
        incompletePath.fill()
    }
}
