//
//  InterstitialView.swift
//  Amalgamation
//
//  Created by Charles Magahern on 5/6/18.
//

import UIKit

class InterstitialView : UIView, Transitionable
{
    private let _imageView: UIImageView = UIImageView()
    private let _label:     UILabel     = UILabel()
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        _imageView.contentMode = .scaleAspectFit
        _imageView.tintColor = UIColor.white
        
        self.addSubview(_imageView)
        self.addSubview(_label)
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: Accessors
    
    var image: UIImage? = nil
    {
        didSet
        {
            _imageView.image = self.image
            self.setNeedsLayout()
        }
    }
    
    var text: String? = nil
    {
        didSet
        {
            let theme = Theme.mainTheme
            let titleAttributes = theme.headerTextAttributes
            let labelAttribString = NSAttributedString(string: (text ?? "").uppercased(), attributes: titleAttributes)
            _label.attributedText = labelAttribString
            
            self.setNeedsLayout()
        }
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        
        let labelSizeThatFits = _label.sizeThatFits(bounds.size)
        let labelFrame = CGRect(
            x: rint(bounds.size.width / 2.0 - labelSizeThatFits.width / 2.0),
            y: bounds.size.height - labelSizeThatFits.height,
            width: labelSizeThatFits.width,
            height: labelSizeThatFits.height
        )
        _label.frame = labelFrame
        
        let imageFrame = CGRect(
            x: 0.0,
            y: 0.0,
            width: bounds.size.width,
            height: bounds.size.height - labelFrame.size.height
        )
        _imageView.frame = imageFrame
    }
    
    // MARK: Transitionable
    
    func transitionToActive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _transitionToActive(active: true, transition: transition, completion: completion)
    }
    
    func transitionToInactive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _transitionToActive(active: false, transition: transition, completion: completion)
    }
    
    // MARK: Internal
    
    func _transitionToActive(active: Bool, transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _imageView.layer.removeAllAnimations()
        _label.layer.removeAllAnimations()
        
        self._imageView.layer.opacity = (active ? 0.0 : 1.0)
        self._label.layer.opacity = (active ? 0.0 : 1.0)
        
        var animations: [CABasicAnimation] = []
        let fadeAnim = CABasicAnimation(keyPath: "opacity")
        fadeAnim.fromValue = (active ? 0.0 : 1.0)
        fadeAnim.toValue = (active ? 1.0 : 0.0)
        animations.append(fadeAnim)

        if active {
            let nonIdentity = CATransform3DMakeTranslation(0.0, (active ? 50.0 : -50.0), 0.0)
            let translateAnim = CABasicAnimation(keyPath: "transform")
            translateAnim.fromValue = nonIdentity
            translateAnim.toValue = CGAffineTransform.identity
            animations.append(translateAnim)
        }
        
        let imageAnimGroup = CAAnimationGroup()
        imageAnimGroup.duration = transition.duration
        imageAnimGroup.animations = animations
        imageAnimGroup.beginTime = CACurrentMediaTime() + transition.delay
        imageAnimGroup.timingFunction = CAMediaTimingFunction(name: (active ? kCAMediaTimingFunctionEaseOut : kCAMediaTimingFunctionEaseIn))
        
        let labelAnimDelay = 0.1
        let labelAnimGroup = imageAnimGroup.copy() as! CAAnimationGroup
        labelAnimGroup.beginTime = labelAnimGroup.beginTime + labelAnimDelay

        CATransaction.begin()
        CATransaction.setCompletionBlock {
            completion(true)
        }
        
        _imageView.layer.add(imageAnimGroup, forKey: nil)
        _label.layer.add(labelAnimGroup, forKey: nil)
        
        CATransaction.commit()
        
        DispatchQueue.main.asyncAfter(deadline: DispatchTime.secondsFromNow(transition.delay)) {
            self._imageView.layer.opacity = (active ? 1.0 : 0.0)
        }
        
        DispatchQueue.main.asyncAfter(deadline: DispatchTime.secondsFromNow(transition.delay + labelAnimDelay)) {
            self._label.layer.opacity = (active ? 1.0 : 0.0)
        }
    }
}
