//
//  TournamentBracketTimelineView.swift
//  Amalgamation
//
//  Created by Charles Magahern on 1/31/18.
//

import Foundation
import UIKit

class TournamentBracketTimelineView : UIView, Transitionable
{
    private var _roundLabels:            [UILabel] = []
    
    private static let _linesPerSegment: Int     = 16
    private static let _majorLineHeight: CGFloat = 13.0
    private static let _minorLineHeight: CGFloat = 5.0
    private static let _lineWidth:       CGFloat = 1.0
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        self.backgroundColor = Theme.mainTheme.backgroundColor
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: Accessors
    
    var roundsToMatchViews: [Int : [TournamentMatchView]] = [:]
    {
        didSet
        {
            self._reloadRoundLabels()
            self.setNeedsDisplay()
        }
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        
        // XXX: this sort function is copied in a bunch of places.
        let sortedRounds = self.roundsToMatchViews.sorted { abs($0.0) < abs($1.0) }
        
        for (roundNumber, matchViews) in sortedRounds {
            if let anyMatchView = matchViews.first, let roundLabel = self.viewWithTag(roundNumber) {
                let matchViewFrame = anyMatchView.frame
                let labelSizeThatFits = roundLabel.sizeThatFits(bounds.size)
                let labelFrame = CGRect(
                    x: rint(matchViewFrame.origin.x + matchViewFrame.size.width / 2.0 - labelSizeThatFits.width / 2.0),
                    y: 0.0,
                    width: labelSizeThatFits.width,
                    height: labelSizeThatFits.height
                )
                roundLabel.frame = labelFrame
            }
        }
    }
    
    override func sizeThatFits(_ size: CGSize) -> CGSize
    {
        var height: CGFloat = 0.0
        
        if let anyLabel = _roundLabels.first {
            height = anyLabel.sizeThatFits(size).height + TournamentBracketTimelineView._majorLineHeight
        }
        
        return CGSize(width: size.width, height: height)
    }
    
    override func draw(_ rect: CGRect)
    {
        super.draw(rect)
        
        let bounds = self.bounds
        let roundsCount = self.roundsToMatchViews.count
        let allMatchViews = _allMatchViews()
        var matchViewsBoundingBox = CGRect.zero
        var matchViewsWidth: CGFloat = 0.0
        
        for matchView in allMatchViews {
            let frame = matchView.frame
            matchViewsBoundingBox.origin.x = min(matchViewsBoundingBox.origin.x, frame.minX)
            matchViewsBoundingBox.origin.y = min(matchViewsBoundingBox.origin.y, frame.minY)
            matchViewsBoundingBox.size.width = max(matchViewsBoundingBox.size.width, frame.maxX)
            matchViewsBoundingBox.size.height = max(matchViewsBoundingBox.size.height, frame.maxY)
            matchViewsWidth = frame.size.width
        }
        
        let matchViewsTotalWidth = matchViewsWidth * CGFloat(roundsCount)
        let matchViewsHMargin = (matchViewsBoundingBox.size.width - matchViewsTotalWidth) / CGFloat(roundsCount - 1)
        let timelineSegmentWidth = matchViewsWidth + matchViewsHMargin
        
        let linesPerSegment = TournamentBracketTimelineView._linesPerSegment
        let majorLineHeight = TournamentBracketTimelineView._majorLineHeight
        let minorLineHeight = TournamentBracketTimelineView._minorLineHeight
        let lineWidth = TournamentBracketTimelineView._lineWidth
        
        let majorLinesOriginY = bounds.size.height - majorLineHeight
        let minorLinesOriginY = majorLinesOriginY + (majorLineHeight / 2.0) - (minorLineHeight / 2.0)
        let lineMargin = (timelineSegmentWidth - (lineWidth * CGFloat(linesPerSegment))) / CGFloat(linesPerSegment)
        
        let path = CGMutablePath()
        
        // add the middle horizontal line first
        let middleLineRect = CGRect(
            x: 0.0,
            y: floor(majorLinesOriginY + majorLineHeight / 2.0),
            width: matchViewsBoundingBox.size.width,
            height: lineWidth
        )
        path.addRect(middleLineRect)
        
        // add the vertical lines for each segment
        for roundNumber in 0..<roundsCount {
            let originX = timelineSegmentWidth * CGFloat(roundNumber) - (matchViewsHMargin / 2.0)
            
            for segmentNumber in 0..<linesPerSegment {
                let isMajorLine = (segmentNumber % (linesPerSegment / 2) == 0)
                let originY = (isMajorLine ? majorLinesOriginY : minorLinesOriginY)
                let height = (isMajorLine ? majorLineHeight : minorLineHeight)

                let rect = CGRect(
                    x: rint(originX + CGFloat(segmentNumber) * (lineWidth + lineMargin)),
                    y: rint(originY),
                    width: rint(lineWidth),
                    height: rint(height)
                )
                path.addRect(rect)
            }
        }
        
        UIColor(white: 0.5, alpha: 1.0).setFill()
        let bezier = UIBezierPath(cgPath: path)
        bezier.fill()
    }
    
    // MARK: Transitionable
    
    func transitionToActive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        UIView.animate(withDuration: transition.duration, delay: transition.delay, options: transition.options, animations: {
            self.alpha = 1.0
        }, completion: completion)
    }
    
    func transitionToInactive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        UIView.animate(withDuration: transition.duration, delay: transition.delay, options: transition.options, animations: {
            self.alpha = 0.0
        }, completion: completion)
    }

    // MARK: Internal
    
    internal func _allMatchViews() -> [TournamentMatchView]
    {
        var allMatchViews: [TournamentMatchView] = []
        
        for (_, matchViews) in self.roundsToMatchViews {
            allMatchViews.append(contentsOf: matchViews)
        }
        
        return allMatchViews
    }
    
    internal func _reloadRoundLabels()
    {
        _roundLabels.forEach { $0.removeFromSuperview() }
        _roundLabels.removeAll()
        
        let theme = Theme.mainTheme        
        let roundNumbers = (self.roundsToMatchViews.reduce(into: [Int]()) { $0.append($1.key) }).sorted()
        let roundOrdinalFormat = NSLocalizedString("ROUND_ORDINAL_FORMAT", comment: "")
        let losersRoundOrdinalFormat = NSLocalizedString("LOSERS_ROUND_ORDINAL_FORMAT", comment: "")
        
        for roundNumber in roundNumbers {
            var labelText: String = ""
            if roundNumber > 0 {
                labelText = String(format: roundOrdinalFormat, roundNumber)
            } else {
                labelText = String(format: losersRoundOrdinalFormat, abs(roundNumber))
            }
            
            let label = UILabel()
            label.attributedText = NSAttributedString(string: labelText.uppercased(), attributes: theme.accessoryTextAttributes)
            label.tag = roundNumber
            
            self.addSubview(label)
            _roundLabels.append(label)
        }
    }
}
