//
//  TournamentBracketView.swift
//  Amalgamation
//
//  Created by Charles Magahern on 1/22/18.
//

import UIKit

struct BracketMaskOptions : OptionSet
{
    let rawValue: Int
    
    static let leftMask  = BracketMaskOptions(rawValue: (1 << 0))
    static let rightMask = BracketMaskOptions(rawValue: (1 << 1))
}

class TournamentBracketView : UIView, UIScrollViewDelegate, Transitionable
{
    private var _matchViewsContainer:   UIView = UIView()
    private var _roundsToMatchViews:    [Int : [TournamentMatchView]] = [:]
    private var _annotationsView:       TournamentBracketAnnotationsView = TournamentBracketAnnotationsView()
    private var _timelineView:          TournamentBracketTimelineView = TournamentBracketTimelineView()
    private var _bracketScrollView:     UIScrollView = UIScrollView()
    private var _bracketViewMaskLayer:  CAGradientLayer = CAGradientLayer()
    private var _active:                Bool = true
    
    fileprivate static let _matchViewsHMargin:  CGFloat = 30.0
    fileprivate static let _matchViewsWidth:    CGFloat = 200.0
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        self.backgroundColor = Theme.mainTheme.backgroundColor
        
        _bracketScrollView.addSubview(_annotationsView)
        _bracketScrollView.addSubview(_timelineView)
        _bracketScrollView.addSubview(_matchViewsContainer)
        _bracketScrollView.delegate = self

        #if DEBUG
            _bracketScrollView.panGestureRecognizer.allowedTouchTypes = [NSNumber(value: UITouchType.indirect.rawValue)]
        #endif

        self.addSubview(_bracketScrollView)
        
        _bracketViewMaskLayer.locations = [0.0, 0.05, 0.95, 1.0]
        _bracketViewMaskLayer.startPoint = CGPoint(x: 0.0, y: 0.5)
        _bracketViewMaskLayer.endPoint = CGPoint(x: 1.0, y: 0.5)
        _bracketViewMaskLayer.rasterizationScale = UIScreen.main.scale
        _bracketViewMaskLayer.shouldRasterize = true

        _setMaskLayerVisible(visible: true, animated: false)
        self._bracketMaskOptions = []
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: API

    var tournament: Tournament?
    {
        didSet
        {
            _reloadBracketViews()
            _reloadBracketMask()
        }
    }
    
    var visibleMatchesRange: CountableClosedRange<Int> = 0...Int.max
    {
        didSet
        {
            _reloadBracketViews()
        }
    }
    
    func scrollToVisibleRound(roundNumber: Int, animated: Bool)
    {
        if let roundMatchViews = _roundsToMatchViews[roundNumber] {
            if let anyRoundMatchView = roundMatchViews[safe: 0] {
                self.layoutIfNeeded()
                
                let anyRoundMatchViewFrame = anyRoundMatchView.frame
                _bracketScrollView.scrollRectToVisible(anyRoundMatchViewFrame, animated: animated)
            }
        }
    }
    
    func scrollToLatestVisibleRound(animated: Bool)
    {
        guard let tournament = self.tournament else { return }
        
        var latestMatch: Match? = nil
        for match in tournament.matches {
            if self.visibleMatchesRange.contains(match.round) {
                if latestMatch == nil && match.dateStarted != nil {
                    latestMatch = match
                } else {
                    guard let latestMatchDate = latestMatch?.dateStarted else { continue }
                    guard let matchDate = match.dateStarted else { continue }
                    if matchDate.compare(latestMatchDate) == .orderedDescending {
                        latestMatch = match
                    }
                }
            }
        }
        
        if let latestMatch = latestMatch {
            self.scrollToVisibleRound(roundNumber: latestMatch.round, animated: animated)
        }
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let timelineHeight = _timelineView.sizeThatFits(bounds.size).height
        let matchViewsWidth = TournamentBracketView._matchViewsWidth
        let matchViewsBounds = CGRect(
            x: bounds.origin.x,
            y: bounds.origin.y + timelineHeight,
            width: bounds.size.width,
            height: bounds.size.height - timelineHeight
        )
        
        // layout parent scroll view
        _bracketScrollView.frame = bounds
        _bracketViewMaskLayer.frame = bounds
        
        // layout each match view by round
        struct RoundLayoutInfo
        {
            var matchCount: Int
            var firstMatchViewFrame: CGRect
            var matchViewVPadding: CGFloat
        }
        
        // XXX: this sort function is copied in a bunch of places.
        let sortedRounds = _roundsToMatchViews.sorted { abs($0.0) < abs($1.0) }
        let hmargin = TournamentBracketView._matchViewsHMargin
        var curOriginX: CGFloat = 0.0
        var maxSize = CGSize.zero
        var previousRoundLayoutInfo: RoundLayoutInfo?
        
        for (_, roundMatchViews) in sortedRounds {
            // calculate the total height of all match views stacked on top of each other
            let matchViewsTotalHeight = roundMatchViews.reduce(into: CGFloat(0.0)) { (result: inout CGFloat, view: TournamentMatchView) in
                result += view.sizeThatFits(matchViewsBounds.size).height
            }
            
            /* if the current round had fewer matches than the previous round, then
             we need to make sure to vertically layout the first match cell in between
             the first two cells of the previous round.  */
            var beginOriginY: CGFloat?
            var availableBoundsHeight = matchViewsBounds.size.height
            var matchViewsCountForVPaddingCalculation = roundMatchViews.count + 1
            
            if let previousRoundLayoutInfo = previousRoundLayoutInfo {
                if previousRoundLayoutInfo.matchCount > roundMatchViews.count {
                    let matchViewFrame = previousRoundLayoutInfo.firstMatchViewFrame
                    beginOriginY = (
                        matchViewFrame.maxY +
                        (previousRoundLayoutInfo.matchViewVPadding / 2.0) -
                        (matchViewFrame.height / 2.0)
                    )
                    availableBoundsHeight = matchViewsBounds.size.height - (2.0 * beginOriginY!)
                    matchViewsCountForVPaddingCalculation = roundMatchViews.count - 1
                }
            }
            
            // calculate vpadding, starting vertical origin, and initialize current round layout
            let vpadding: CGFloat = (availableBoundsHeight - matchViewsTotalHeight) / CGFloat(matchViewsCountForVPaddingCalculation)
            var curOriginY: CGFloat = beginOriginY ?? vpadding
            var curRoundLayoutInfo = RoundLayoutInfo(
                matchCount: roundMatchViews.count,
                firstMatchViewFrame: CGRect.zero,
                matchViewVPadding: vpadding
            )
            
            // layout each match view
            for (matchIndex, matchView) in roundMatchViews.enumerated() {
                let sizeThatFits = matchView.sizeThatFits(matchViewsBounds.size)
                let matchViewFrame = CGRect(
                    x: rint(curOriginX),
                    y: rint(curOriginY),
                    width: rint(matchViewsWidth),
                    height: rint(sizeThatFits.height)
                )
                matchView.frame = matchViewFrame
                
                curOriginY += matchViewFrame.height + vpadding
                maxSize.width = max(maxSize.width, matchViewFrame.maxX)
                maxSize.height = max(maxSize.height, matchViewFrame.maxY)
                
                if matchIndex == 0 {
                    curRoundLayoutInfo.firstMatchViewFrame = matchViewFrame
                }
            }
            
            // update layout info for the next iteration
            curOriginX += matchViewsWidth + hmargin
            previousRoundLayoutInfo = curRoundLayoutInfo
        }
        
        // update scroll view properties
        _bracketScrollView.contentSize = maxSize
        _bracketScrollView.contentInset = UIEdgeInsets(
            top: 0.0,
            left: bounds.size.width / 2.0 - maxSize.width / 2.0,
            bottom: 0.0,
            right: 0.0
        )
        
        // update timeline view frame
        let timelineFrame = CGRect(
            x: bounds.origin.x,
            y: bounds.origin.y,
            width: maxSize.width,
            height: timelineHeight
        )
        _timelineView.frame = timelineFrame
        
        // update annotations view frame
        _matchViewsContainer.frame = CGRect(origin: matchViewsBounds.origin, size: maxSize)
        _annotationsView.frame = CGRect(origin: matchViewsBounds.origin, size: maxSize)
    }
    
    override var canBecomeFocused: Bool
    {
        #if DEBUG
            return true
        #else
            return false
        #endif
    }
    
    // MARK: UIScrollViewDelegate
    
    func scrollViewDidScroll(_ scrollView: UIScrollView)
    {
        if scrollView != _bracketScrollView { return }
        _reloadBracketMask()
    }
    
    // MARK: Transitionable
    
    func transitionToActive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _transitionToActive(active: true, transition: transition, completion: completion)
    }
    
    func transitionToInactive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _transitionToActive(active: false, transition: transition, completion: completion)
    }
    
    // MARK: Internal
    
    internal var _bracketMaskOptions: BracketMaskOptions = [.rightMask]
    {
        didSet
        {
            if oldValue != _bracketMaskOptions {
                var leftEdgeColor = UIColor.white
                var rightEdgeColor = UIColor.white
                
                if _bracketMaskOptions.contains(.leftMask) {
                    leftEdgeColor = UIColor.clear
                }
                
                if _bracketMaskOptions.contains(.rightMask) {
                    rightEdgeColor = UIColor.clear
                }
                
                _bracketViewMaskLayer.colors = [
                    leftEdgeColor.cgColor,
                    UIColor.white.cgColor,
                    UIColor.white.cgColor,
                    rightEdgeColor.cgColor
                ]
            }
        }
    }
    
    internal func _reloadBracketMask()
    {
        let scrollViewBounds = _bracketScrollView.bounds
        let scrollViewContentSize = _bracketScrollView.contentSize
        let scrollViewOffset = _bracketScrollView.contentOffset
        
        var newBracketMaskOptions = BracketMaskOptions()
        if scrollViewOffset.x > CGFloat.leastNonzeroMagnitude {
            newBracketMaskOptions.insert(.leftMask)
        }
        
        if scrollViewContentSize.width - scrollViewOffset.x > scrollViewBounds.size.width {
            newBracketMaskOptions.insert(.rightMask)
        }

        self._bracketMaskOptions = newBracketMaskOptions
    }
    
    internal func _reloadBracketViews()
    {
        _roundsToMatchViews.values.forEach { $0.forEach { $0.removeFromSuperview() } }
        _roundsToMatchViews.removeAll()
        
        if let tournament = self.tournament {
            let matches = tournament.matches
            
            for match in matches {
                guard let player1ID = match.player1ID else { continue }
                guard let player2ID = match.player2ID else { continue }
                let round = match.round

                if self.visibleMatchesRange.contains(round) {
                    let player1 = tournament.participantsMap[player1ID]
                    let player2 = tournament.participantsMap[player2ID]
                    let tournamentMatch = TournamentMatch(match: match, player1Participant: player1, player2Participant: player2)
                    
                    let matchView = TournamentMatchView()
                    matchView.match = tournamentMatch
                    matchView.alpha = _active ? 1.0 : 0.0
                    
                    var matchViews = _roundsToMatchViews[round] ?? [TournamentMatchView]()
                    matchViews.append(matchView)
                    
                    _matchViewsContainer.addSubview(matchView)
                    _roundsToMatchViews[round] = matchViews
                }
            }
        }
        
        _timelineView.roundsToMatchViews = _roundsToMatchViews
        _annotationsView.roundsToMatchViews = _roundsToMatchViews
        
        self.setNeedsLayout()
    }
    
    internal func _setMaskLayerVisible(visible: Bool, animated: Bool)
    {
        let addOrRemoveLayerBlock = {
            if visible {
                self.layer.addSublayer(self._bracketViewMaskLayer)
                self.layer.mask = self._bracketViewMaskLayer
            } else {
                self._bracketViewMaskLayer.removeFromSuperlayer()
                self.layer.mask = nil
            }
        }
        
        if animated {
            if !visible {
                self._bracketMaskOptions = []
            } else {
                _reloadBracketMask()
            }
            
            addOrRemoveLayerBlock()
        } else {
            addOrRemoveLayerBlock()
        }
    }
    
    internal func _transitionToActive(active: Bool, transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _active = active
        
        // disable layer mask for animation performance
        _setMaskLayerVisible(visible: false, animated: active)

        // setup dispatch group and subview animation callbacks
        let group = DispatchGroup()
        var groupComplete = false
        let subviewTransitionCompletion = { (complete: Bool) in
            groupComplete = (groupComplete && complete)
            group.leave()
        }

        // transition timeline view
        group.enter()
        if active {
            self._timelineView.transitionToActive(transition: transition, completion: subviewTransitionCompletion)
        } else {
            self._timelineView.transitionToInactive(transition: transition, completion: subviewTransitionCompletion)
        }

        // transition individual match views
        let sortedRounds = self._roundsToMatchViews.sorted { abs($0.0) < abs($1.0) }
        let matchViewsCount = self._roundsToMatchViews.reduce(0) { $0 + $1.value.count }
        let matchViewDelayJitter = (500.0 / Double(matchViewsCount)) / 1000.0
        var curMatchViewDelay = transition.delay

        var matchViewTransition = transition
        matchViewTransition.duration = transition.duration - (TimeInterval(matchViewsCount - 1) * matchViewDelayJitter)

        for (_, roundMatchViews) in sortedRounds {
            for matchView in roundMatchViews {
                matchViewTransition.delay = curMatchViewDelay

                group.enter()
                if active {
                    matchView.transitionToActive(transition: matchViewTransition, completion: subviewTransitionCompletion)
                } else {
                    matchView.transitionToInactive(transition: matchViewTransition, completion: subviewTransitionCompletion)
                }

                curMatchViewDelay += matchViewDelayJitter
            }
        }

        // transition annotations view
        var annotationsTransition = transition
        annotationsTransition.duration = transition.duration / 4.0

        group.enter()
        if active {
            annotationsTransition.delay = curMatchViewDelay
            self._annotationsView.transitionToActive(transition: annotationsTransition, completion: subviewTransitionCompletion)
        } else {
            annotationsTransition.delay = transition.delay
            self._annotationsView.transitionToInactive(transition: annotationsTransition, completion: subviewTransitionCompletion)
        }

        group.notify(queue: DispatchQueue.main) {
            self._setMaskLayerVisible(visible: true, animated: true)
            completion(groupComplete)
        }
    }
}
