//
//  TournamentMatchView.swift
//  Amalgamation
//
//  Created by Charles Magahern on 1/22/18.
//

import UIKit

class TournamentMatchView : UIView, Transitionable
{
    private var _player1View:   MatchPlayerView = MatchPlayerView()
    private var _player2View:   MatchPlayerView = MatchPlayerView()
    
    private static let _playerViewsVMargin: CGFloat = 1.0
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        self.addSubview(_player1View)
        self.addSubview(_player2View)
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: API
    
    var match: TournamentMatch?
    {
        didSet
        {
            let player1 = match?.player1Participant
            _player1View.seed = player1?.seed
            _player1View.playerName = player1?.name
            _player1View.score = match?.match.player1Score
            
            let player2 = match?.player2Participant
            _player2View.seed = player2?.seed
            _player2View.playerName = player2?.name
            _player2View.score = match?.match.player2Score
            
            if let winnerID = match?.match.winnerID {
                _player1View.winner = (player1?.id == winnerID)
                _player2View.winner = (player2?.id == winnerID)
            }
        }
    }
    
    // MARK: UIView
    
    override func sizeThatFits(_ size: CGSize) -> CGSize
    {
        let player1Size = _player1View.sizeThatFits(size)
        let player2Size = _player2View.sizeThatFits(size)
        return CGSize(
            width: max(player1Size.width, player2Size.width),
            height: player1Size.height + player2Size.height + TournamentMatchView._playerViewsVMargin
        )
    }
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let playerViewHeight: CGFloat = rint(bounds.size.height / 2.0)
        
        let player1ViewFrame = CGRect(
            x: 0.0,
            y: 0.0,
            width: bounds.size.width,
            height: playerViewHeight
        )
        _player1View.frame = player1ViewFrame
        
        let player2ViewFrame = CGRect(
            x: 0.0,
            y: player1ViewFrame.maxY + TournamentMatchView._playerViewsVMargin,
            width: bounds.size.width,
            height: playerViewHeight
        )
        _player2View.frame = player2ViewFrame
    }
    
    // MARK: Transitionable
    
    func transitionToActive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _transitionToActive(active: true, transition: transition, completion: completion)
    }
    
    func transitionToInactive(transition: Transition, completion: @escaping (Bool) -> Void)
    {
        _transitionToActive(active: false, transition: transition, completion: completion)
    }
    
    // MARK: Internal
    
    internal func _transitionToActive(active: Bool, transition: Transition, completion: @escaping (Bool) -> Void)
    {
        let beginAlpha = CGFloat(active ? 0.0 : 1.0)
        let destAlpha = CGFloat(active ? 1.0 : 0.0)
        
        let nonIdentity = CGAffineTransform(translationX: (active ? 50.0 : -50.0), y: (active ? -10.0 : 10.0))

        let beginTransform = (active ? nonIdentity : CGAffineTransform.identity)
        let destTransform = (active ? CGAffineTransform.identity : nonIdentity)
        
        self.alpha = beginAlpha
        self.transform = beginTransform
        
        UIView.animate(withDuration: transition.duration, delay: transition.delay, options: transition.options, animations: {
            self.alpha = destAlpha
            self.transform = destTransform
        }, completion: completion)
    }
}

internal class MatchPlayerView : UIView
{
    private var _seedLabel:                         UILabel = UILabel()
    private var _playerNameLabel:                   PaddedLabel = PaddedLabel()
    private var _pointsEarnedLabel:                 UILabel = UILabel()
    
    private var _numericalLabelColor:               UIColor
    private var _numericalLabelHighlightedColor:    UIColor
    private var _textualLabelColor:                 UIColor
    
    private static let _numericalLabelBackgroundColor = UIColor(red: 0.66, green: 0.66, blue: 0.66, alpha: 1.00)
    private static let _numericalLabelHighlightedColor = UIColor(red: 0.79, green: 0.66, blue: 0.15, alpha: 1.00)
    private static let _textualLabelBackgroundColor = UIColor(red: 0.15, green: 0.20, blue: 0.23, alpha: 1.00)
    
    override init(frame: CGRect)
    {
        let theme = Theme.mainTheme
        _numericalLabelColor = UIColor(red: 0.66, green: 0.66, blue: 0.66, alpha: 1.00)
        _numericalLabelHighlightedColor = theme.colorPalette.primaryColor
        _textualLabelColor = theme.colorPalette.secondaryColor
        
        super.init(frame: frame)
        
        let accessoryLabelTextColor = UIColor(white: 0.2, alpha: 1.0)
        let font = theme.accessoryTextAttributes[NSAttributedStringKey.font] as! UIFont
        
        _seedLabel.backgroundColor = _numericalLabelColor
        _seedLabel.textColor = accessoryLabelTextColor
        _seedLabel.textAlignment = .center
        _seedLabel.font = font
        self.addSubview(_seedLabel)

        _playerNameLabel.leftInset = 5.0
        _playerNameLabel.rightInset = 5.0
        _playerNameLabel.backgroundColor = _textualLabelColor
        _playerNameLabel.textColor = theme.foregroundColor
        _playerNameLabel.textAlignment = .left
        _playerNameLabel.font = font
        self.addSubview(_playerNameLabel)
        
        _pointsEarnedLabel.backgroundColor = _numericalLabelColor
        _pointsEarnedLabel.textColor = accessoryLabelTextColor
        _pointsEarnedLabel.textAlignment = .center
        _pointsEarnedLabel.font = font
        self.addSubview(_pointsEarnedLabel)
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: API
    
    var playerName: String?
    {
        didSet
        {
            _playerNameLabel.text = playerName?.uppercased()
            self.setNeedsLayout()
        }
    }
    
    var seed: Int?
    {
        didSet
        {
            if let seed = self.seed {
                _seedLabel.text = String(seed)
            } else {
                _seedLabel.text = "?"
            }
            
            self.setNeedsLayout()
        }
    }
    
    var score: Int?
    {
        didSet
        {
            if let score = self.score {
                _pointsEarnedLabel.text = String(score)
                _pointsEarnedLabel.isHidden = false
            } else {
                _pointsEarnedLabel.isHidden = true
            }
        }
    }
    
    var winner: Bool?
    {
        didSet
        {
            if winner == true {
                _pointsEarnedLabel.backgroundColor = _numericalLabelHighlightedColor
            } else {
                _pointsEarnedLabel.backgroundColor = _numericalLabelColor
            }
        }
    }
    
    // MARK: UIView
    
    override func sizeThatFits(_ size: CGSize) -> CGSize
    {
        let labels = [_seedLabel, _playerNameLabel, _pointsEarnedLabel]
        let labelsMaxHeight = labels.reduce(into: CGFloat(0.0)) { (result: inout CGFloat, label: UIView) in
            result = max(result, label.sizeThatFits(size).height)
        }
        
        return CGSize(width: size.width, height: labelsMaxHeight + 5.0)
    }
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let accessoryLabelsWidth: CGFloat = 25.0
        let accessoryLabels = [_seedLabel, _pointsEarnedLabel]
        let visibleAccessoryLabelsWidth = accessoryLabels.reduce(into: CGFloat(0.0)) { (result: inout CGFloat, label: UILabel) in
            result += (label.isHidden ? 0.0 : accessoryLabelsWidth)
        }
        
        let seedLabelFrame = CGRect(
            x: 0.0,
            y: 0.0,
            width: accessoryLabelsWidth,
            height: bounds.size.height
        )
        _seedLabel.frame = seedLabelFrame
        
        let playerNameContainerFrame = CGRect(
            x: seedLabelFrame.maxX,
            y: 0.0,
            width: bounds.size.width - visibleAccessoryLabelsWidth,
            height: bounds.size.height
        )
        _playerNameLabel.frame = playerNameContainerFrame

        let pointsEarnedLabelFrame = CGRect(
            x: playerNameContainerFrame.maxX,
            y: 0.0,
            width: accessoryLabelsWidth,
            height: bounds.size.height
        )
        _pointsEarnedLabel.frame = pointsEarnedLabelFrame
    }
}
