//
//  TournamentStateView.swift
//  Amalgamation
//
//  Created by Charles Magahern on 1/15/18.
//

import QuartzCore
import UIKit

class TournamentStateView : UIView
{
    private var _indicator:         TournamentStateIndicator = TournamentStateIndicator()
    private var _stateLabel:        UILabel = UILabel()
    private var _timeElapsedLabel:  UILabel = UILabel()
    private var _updateTimer:       Timer?
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        self.addSubview(_indicator)
        self.addSubview(_stateLabel)
        self.addSubview(_timeElapsedLabel)
        
        _updateTournamentState()
        _updateTimeElapsedLabel()
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    deinit
    {
        _updateTimer?.invalidate()
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let indicatorFrame = CGRect(
            x: 0.0,
            y: 0.0,
            width: bounds.size.height,
            height: bounds.size.height
        )
        _indicator.frame = indicatorFrame
        
        let labelsVPadding: CGFloat = 5.0
        let stateLabelSizeThatFits = _stateLabel.sizeThatFits(bounds.size)
        let timeElapsedLabelSizeThatFits = _timeElapsedLabel.sizeThatFits(bounds.size)
        let totalLabelsHeight = (stateLabelSizeThatFits.height + labelsVPadding + timeElapsedLabelSizeThatFits.height)
        
        let stateLabelFrame = CGRect(
            x: indicatorFrame.maxX + 10.0,
            y: rint(indicatorFrame.origin.y + (bounds.size.height / 2.0 - totalLabelsHeight / 2.0)),
            width: stateLabelSizeThatFits.width,
            height: stateLabelSizeThatFits.height
        )
        _stateLabel.frame = stateLabelFrame
        
        let timeElapsedFrame = CGRect(
            x: stateLabelFrame.origin.x,
            y: stateLabelFrame.maxY + labelsVPadding,
            width: timeElapsedLabelSizeThatFits.width,
            height: timeElapsedLabelSizeThatFits.height
        )
        _timeElapsedLabel.frame = timeElapsedFrame
    }
    
    // MARK: API
    
    var tournament: Tournament?
    {
        didSet
        {
            _updateTournamentState()
            _updateTimeElapsedLabel()
        }
    }
    
    var animating: Bool = false
    {
        didSet
        {
            _indicator.animating = animating
            
            if animating {
                _updateTimer = Timer.scheduledTimer(withTimeInterval: 0.5, repeats: true, block: { [weak self] (timer: Timer) in
                    self?._updateTimeElapsedLabel()
                })
            } else {
                _updateTimer?.invalidate()
                _updateTimer = nil
            }
        }
    }
    
    // MARK: Internal
    
    internal func _updateTournamentState()
    {
        let state = self.tournament?.state ?? .unknown
        
        let localizedStateName = state.localizedDescription
        _stateLabel.attributedText = NSAttributedString(string: localizedStateName.uppercased(),
                                                        attributes: Theme.mainTheme.statusTextAttributes)
        
        _indicator.tournamentState = state
        
        self.setNeedsLayout()
    }
    
    internal func _updateTimeElapsedLabel()
    {
        var intervalElapsed: TimeInterval = 0.0
        
        if let tournament = self.tournament {
            switch tournament.state {
            case .pending:
                intervalElapsed = 0.0
            case .underway:
                if let dateStarted = tournament.dateStarted {
                    intervalElapsed = dateStarted.timeIntervalSinceNow
                }
            case .complete:
                if let dateStarted = tournament.dateStarted, let dateCompleted = tournament.dateCompleted {
                    intervalElapsed = dateCompleted.timeIntervalSince(dateStarted)
                }
            default:
                intervalElapsed = 0.0
            }
        }
        
        let attributes = Theme.mainTheme.statusTextAttributes
        let seconds = UInt(abs(intervalElapsed))
        let intervalString = String(format: "%02u:%02u:%02u", seconds / 3600, (seconds / 60) % 60, seconds % 60)
        _timeElapsedLabel.attributedText = NSAttributedString(string: intervalString, attributes: attributes)
    }
}

internal class TournamentStateIndicator : UIView
{
    private var _leftRect:      CAGradientLayer = CAGradientLayer()
    private var _rightRect:     CAGradientLayer = CAGradientLayer()
    private var _bottomRect:    CAGradientLayer = CAGradientLayer()
    
    override init(frame: CGRect)
    {
        super.init(frame: frame)
        
        let layer = self.layer

        _bottomRect.startPoint = CGPoint(x: 0.0, y: 0.5)
        _bottomRect.endPoint = CGPoint(x: 1.0, y: 0.5)
        
        // gradient is symmetric, but reverse the right side's gradient for
        // animation purposes
        _rightRect.startPoint = CGPoint(x: 0.5, y: 1.0)
        _rightRect.endPoint = CGPoint(x: 0.5, y: 0.0)

        layer.addSublayer(_leftRect)
        layer.addSublayer(_rightRect)
        layer.addSublayer(_bottomRect)
    }
    
    required init?(coder aDecoder: NSCoder)
    {
        fatalError("unsupported")
    }
    
    // MARK: UIView
    
    override func layoutSubviews()
    {
        super.layoutSubviews()
        
        let bounds = self.bounds
        let rectHeight: CGFloat = 12.0
        let rectWidth = bounds.size.width - (rectHeight * 2.0)
        let θ = CGFloat.pi / 6.0
        let triangleHeight = rint(rectHeight + rectWidth * cos(θ) + rectHeight * sin(θ))

        // bottom rect
        let bottomRect = CGRect(
            x: rectHeight,
            y: rint(bounds.size.height / 2.0 - triangleHeight / 2.0) + (triangleHeight - rectHeight),
            width: rectWidth,
            height: rectHeight
        )
        _bottomRect.anchorPoint = CGPoint(x: 0.0, y: 0.0)
        _bottomRect.position = bottomRect.origin
        _bottomRect.bounds = CGRect(origin: CGPoint.zero, size: bottomRect.size)

        // left rect
        let leftRectOrigin = bottomRect.origin
        _leftRect.anchorPoint = CGPoint(x: 1.0, y: 1.0)
        _leftRect.position = leftRectOrigin
        _leftRect.bounds = CGRect(x: 0.0, y: 0.0, width: rectHeight, height: rectWidth)
        _leftRect.transform = CATransform3DMakeRotation(CGFloat.pi / 6.0, 0.0, 0.0, 1.0)

        // right rect
        let rightRectOrigin = CGPoint(
            x: bounds.size.width - rectHeight,
            y: bottomRect.origin.y
        )
        _rightRect.anchorPoint = CGPoint(x: 0.0, y: 1.0)
        _rightRect.position = rightRectOrigin
        _rightRect.bounds = CGRect(x: 0.0, y: 0.0, width: rectHeight, height: rectWidth)
        _rightRect.transform = CATransform3DMakeRotation(-CGFloat.pi / 6.0, 0.0, 0.0, 1.0)
    }
    
    // MARK: API
    
    var animating: Bool = false
    {
        didSet
        {
            if oldValue != animating {
                _reloadAnimationState()
            }
        }
    }
    
    var tournamentState: Tournament.State = .unknown
    {
        didSet
        {
            if oldValue != tournamentState {
                _reloadAnimationState()
            }
        }
    }
    
    // MARK: Internal
    
    internal func _reloadRectGradientColors()
    {
        let theme = Theme.mainTheme
        let mainColor = theme.foregroundColor
        var rectColors: [CGColor]
        
        if self.animating && self.tournamentState == .underway {
            let edgeColor = mainColor.withAlphaComponent(0.5)
            rectColors = [
                edgeColor.cgColor, // far edge
                edgeColor.cgColor, // > begin window
                mainColor.cgColor, // - middle window
                edgeColor.cgColor, // < end window
                edgeColor.cgColor  // far edge
            ]
        } else {
            rectColors = [
                mainColor.cgColor,
                mainColor.cgColor,
                mainColor.cgColor,
                mainColor.cgColor,
                mainColor.cgColor
            ]
        }
        
        for rect in [_leftRect, _rightRect, _bottomRect] {
            rect.colors = rectColors
            rect.shouldRasterize = true
            rect.rasterizationScale = UIScreen.main.scale
        }
    }
    
    internal func _reloadAnimationState()
    {
        _reloadRectGradientColors()
        
        _leftRect.removeAllAnimations()
        _rightRect.removeAllAnimations()
        _bottomRect.removeAllAnimations()
        
        if self.animating {
            switch self.tournamentState {
            case .pending:
                let pulseAnimation = CABasicAnimation(keyPath: "opacity")
                pulseAnimation.fromValue = 0.4
                pulseAnimation.toValue = 1.0
                pulseAnimation.duration = 1.5
                pulseAnimation.repeatCount = Float.greatestFiniteMagnitude
                pulseAnimation.autoreverses = true
                pulseAnimation.timingFunction = CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseInEaseOut)
                
                _leftRect.add(pulseAnimation, forKey: nil)
                _rightRect.add(pulseAnimation, forKey: nil)
                _bottomRect.add(pulseAnimation, forKey: nil)
            
            case .underway:
                let duration = 7.0
                let skew = 4.0
                let gradientWidth = 4.0
                let gradientBounds = 1.0 + gradientWidth
                let startLocations = [-gradientBounds,
                                      -gradientBounds + gradientWidth / 4.0,
                                      -gradientBounds + gradientWidth / 2.0,
                                      -gradientBounds + gradientWidth,
                                      gradientBounds]
                let endLocations = [-gradientBounds,
                                    gradientBounds - gradientWidth,
                                    gradientBounds - gradientWidth / 2.0,
                                    gradientBounds - gradientWidth / 4.0,
                                    gradientBounds]
                
                let gradientAnim = CABasicAnimation(keyPath: "locations")
                gradientAnim.fromValue = startLocations
                gradientAnim.toValue = endLocations
                gradientAnim.duration = (duration / 3.0)
                gradientAnim.autoreverses = false
                
                // bottom rect animation
                let bottomAnimGroup = CAAnimationGroup()
                bottomAnimGroup.duration = duration - skew
                bottomAnimGroup.repeatCount = Float.greatestFiniteMagnitude
                bottomAnimGroup.animations = [gradientAnim]

                _bottomRect.locations = startLocations as [NSNumber]
                _bottomRect.add(bottomAnimGroup, forKey: nil)

                // right rect animation
                let rightAnimGroup = bottomAnimGroup.copy() as! CAAnimationGroup
                rightAnimGroup.beginTime = CACurrentMediaTime() + ((duration - skew) / 3.0)

                _rightRect.locations = startLocations as [NSNumber]
                _rightRect.add(rightAnimGroup, forKey: nil)

                // left rect animation
                let leftAnimGroup = bottomAnimGroup.copy() as! CAAnimationGroup
                leftAnimGroup.beginTime = CACurrentMediaTime() + 2.0 * ((duration - skew) / 3.0)

                _leftRect.locations = startLocations as [NSNumber]
                _leftRect.add(leftAnimGroup, forKey: nil)
            default:
                break
            }
        }
    }
}
